import os
import sys

from pysitk.definitions import DIR_TMP

DIR_ROOT = os.path.realpath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
DIR_TEST = os.path.join(DIR_ROOT, "data")

# Constants
EPS = 1e-10

REGEX_FILENAMES = "[A-Za-z0-9+-_]+"
FILENAME_EXTENSION = "txt"

ALLOWED_INPUT_FILE_EXTENSIONS = ["mat", "png", "nii", "nii.gz"]
ALLOWED_NOISE_TYPES = ["gaussian", "poisson", "s&p", "uniform"]



# \file DataWriter.py
#  \brief Writes numpy arrays to files
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date September 2017


# Import libraries
import os
import sys
import SimpleITK as sitk
import scipy.io
import numpy as np
import re
import natsort

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh


class DataWriter(object):

    ##
    # Store relevant reader information
    # \date       2017-09-13 12:00:29+0100
    #
    # \param      self          The object
    # \param      nda           numpy data array
    # \param      path_to_file  Path to write image
    # \param      image_sitk    sitk.Image object for (optional) nii header in
    #                           case of writing to nii/nii.gz file
    #
    def __init__(self, nda, path_to_file, image_sitk=None):
        self._nda = nda
        self._path_to_file = path_to_file
        self._image_sitk = image_sitk

        # Get filename extension
        self._file_type = os.path.basename(self._path_to_file).split(".")[1]

        self._write_data = {
            "txt": self._write_data_txt,
            "png": self._write_data_png,
            "mat": self._write_data_mat,
            "nii": self._write_data_nii,
        }

    def write_data(self):
        ph.create_directory(os.path.dirname(self._path_to_file))
        self._write_data[self._file_type]()
        ph.print_info("File written to '%s'" % self._path_to_file)

    def _write_data_png(self):
        nda = np.round(np.array(self._nda)).astype(np.uint8)
        ph.write_image(nda, self._path_to_file)

    def _write_data_txt(self):
        ph.write_array_to_file(self._path_to_file, self._nda)

    def _write_data_mat(self):
        dic = {"nda": self._nda}
        scipy.io.savemat(self._path_to_file, dic)
        ph.print_info("File written to '%s'" % (self._path_to_file))

    def _write_data_nii(self):
        image_sitk = sitk.GetImageFromArray(self._nda)

        if self._image_sitk is not None:
            image_sitk.CopyInformation(self._image_sitk)

        sitkh.write_nifti_image_sitk(image_sitk, self._path_to_file)



##
# \file prior_measures.py
# \brief      Collection of prior measures, i.e. regularizer costs
#
# \todo proper unit testing of these functions to be done yet
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#


import numpy as np

from nsol.loss_functions import LossFunctions as loss_fun


class PriorMeasures(object):

    @staticmethod
    def zeroth_order_tikhonov(x):
        return 0.5 * np.sum(np.square(x))

    @staticmethod
    def first_order_tikhonov(x, D):
        return 0.5 * np.sum(np.square(D(x)))

    @staticmethod
    def total_variation(x, D, dimension):
        Dx_split = np.array_split(D(x), dimension)

        # Dx_i ** 2
        sum_Dx_i_squared = Dx_split[0]**2

        # Compute sum_k x_k^2
        for i in range(1, len(Dx_split)):
            sum_Dx_i_squared += Dx_split[i]**2

        return np.sum(np.sqrt(sum_Dx_i_squared))

    @staticmethod
    def huber(x, D, dimension, gamma=0.05):
        Dx_split = np.array_split(D(x), dimension)

        # Dx_i ** 2
        sum_Dx_i_squared = Dx_split[0]**2

        # Compute sum_k x_k^2
        for i in range(1, len(Dx_split)):
            sum_Dx_i_squared += Dx_split[i]**2

        Dx_gamma = loss_fun.huber(sum_Dx_i_squared, gamma=gamma) / (2.*gamma)

        return np.sum(Dx_gamma)



##
# \file kernels.py
# \brief      Class to create kernels used for blurring and differentiation in
#             1D, 2D and 3D
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

# Import libraries
import numpy as np
from abc import ABCMeta, abstractmethod


class Kernels(object):
    __metaclass__ = ABCMeta

    def __init__(self, dimension, spacing):

        spacing = np.atleast_1d(spacing).astype(float)

        if spacing.size != dimension:
            raise ValueError("dimension of spacing and space must be the same")

        self._dimension = dimension
        self._spacing = spacing

    def get_dimension(self):
        return self._dimension

    def get_spacing(self):
        return self._spacing

    ##
    # Gets the Gaussian kernel created from a given covariance matrix and
    # cut-off distance.
    # \date       2017-07-19 12:48:06+0100
    #
    # \param      self       The object
    # \param      cov        Variance covariance matrix as numpy array
    # \param      alpha_cut  Cut-off distance in integer, i.e. 3 means cutting
    #                        off a 3 sigma in each direction
    # \param      spacing    Image spacing in x-, y- (z-) direction as numpy
    #                        array
    #
    # \return     The Gaussian kernel with its size depending on alpha_cut
    #
    @abstractmethod
    def get_gaussian(self):
        pass

    ##
    # Get kernel to compute derivative in x-direction based on forward
    # difference.
    #
    # Kernel is assumed to be applied on arrays of the form array =
    # array[(z,)y,x], i.e. the 'correct' direction by viewing the resulting
    # nifti-image differentiation. Resulting kernel can be used via
    # ndimage.convolve(nda, kernel)
    # \date       2017-07-19 15:13:04+0100
    #
    # \param      self     The object
    #
    # \return     kernel to differentiate array in x-direction.
    #
    @abstractmethod
    def get_dx_forward_difference(self):
        pass

    @abstractmethod
    def get_dx_backward_difference(self):
        pass


class Kernels1D(Kernels):

    def __init__(self, spacing=1):
        super(self.__class__, self).__init__(dimension=1, spacing=spacing)

    def get_gaussian(self, cov, alpha_cut=3):

        # Generate intervals for x based on cut-off distance given by
        # standard deviation, alpha_cut and spacing
        x_max = np.ceil(np.sqrt(cov) * alpha_cut/self._spacing)

        step = 1
        points = np.arange(-x_max, x_max+step, step)

        # Mean/Origin for Gaussian blurring
        origin = 0.

        # Compute scaled, inverse covariance matrix
        cov_scale_inv = self._spacing**2 / cov

        # Compute Gaussian weights
        values = (points-origin)*cov_scale_inv * (points-origin)
        kernel = np.exp(-0.5 * values)
        kernel = kernel/np.sum(kernel)

        return kernel

    def get_dx_forward_difference(self):
        kernel = np.zeros(2)
        kernel = np.array([1, -1])

        return kernel / self._spacing

    def get_dx_backward_difference(self):
        kernel = np.zeros(3)
        kernel = np.array([0, 1, -1])

        return kernel / self._spacing


class Kernels2D(Kernels):

    def __init__(self, spacing=np.ones(2)):
        super(self.__class__, self).__init__(dimension=2, spacing=spacing)

    def get_gaussian(self, cov, alpha_cut=3):

        if cov.shape != (self._dimension, self._dimension):
            raise ValueError("Numpy array 'cov' must be of shape (%d,%d)" %
                             (self._dimension, self._dimension))

        # Generate intervals for x and y based on cut-off distance given by
        # standard deviation, alpha_cut and spacing
        [x_max, y_max] = np.ceil(
            np.sqrt(cov.diagonal()) * alpha_cut/self._spacing)

        step = 1
        x_interval = np.arange(-x_max, x_max+step, step)
        y_interval = np.arange(-y_max, y_max+step, step)

        # Generate arrays of 2D points bearing in mind that nifti-nda.shape =
        # (y,x)-coord. 'ij' yields vertical x-coordinate for image
        [X, Y] = np.meshgrid(x_interval, y_interval,
                             indexing='ij')
        points = np.array([Y.flatten(), X.flatten()])

        # Mean/Origin for Gaussian blurring
        origin = np.zeros((self._dimension, 1))

        # Scaling matrix depending on spacing
        S = np.diag(self._spacing)

        # Compute scaled, inverse covariance matrix
        cov_scale_inv = S.dot(np.linalg.inv(cov)).dot(S)

        # Compute Gaussian weights
        values = np.sum((points-origin)*cov_scale_inv.dot(points-origin), 0)
        kernel = np.exp(-0.5 * values)
        kernel = kernel/np.sum(kernel)

        # Reshape kernel
        kernel = kernel.reshape(x_interval.size, y_interval.size)

        return kernel

    def get_dx_forward_difference(self):

        # kernel = np.zeros((y,x))
        kernel = np.zeros((1, 2))
        kernel[:] = np.array([1, -1])

        return kernel / self._spacing[0]

    def get_dx_backward_difference(self):

        # kernel = np.zeros((y,x))
        kernel = np.zeros((1, 3))
        kernel[:] = np.array([0, 1, -1])

        return kernel / self._spacing[0]

    def get_dy_forward_difference(self):

        # kernel = np.zeros((y,x))
        kernel = np.zeros((2, 1))
        kernel[:] = np.array([[1], [-1]])

        return kernel / self._spacing[1]

    def get_dy_backward_difference(self):

        # kernel = np.zeros((y,x))
        kernel = np.zeros((3, 1))
        kernel[:] = np.array([[0], [1], [-1]])

        return kernel / self._spacing[1]


class Kernels3D(Kernels):

    def __init__(self, spacing=np.ones(3)):
        super(self.__class__, self).__init__(dimension=3, spacing=spacing)

    def get_gaussian(self, cov, alpha_cut=3):

        if cov.shape != (self._dimension, self._dimension):
            raise ValueError("Numpy array 'cov' must be of shape (%d,%d)" %
                             (self._dimension, self._dimension))

        # Generate intervals for x, y and z based on cut-off distance given by
        # standard deviation, alpha_cut and spacing
        [x_max, y_max, z_max] = np.ceil(
            np.sqrt(cov.diagonal()) * alpha_cut/self._spacing)

        step = 1
        x_interval = np.arange(-x_max, x_max+step, step)
        y_interval = np.arange(-y_max, y_max+step, step)
        z_interval = np.arange(-z_max, z_max+step, step)

        # Generate arrays of 3D points bearing in mind that nifti-nda.shape =
        # (z,y,x)-coord. 'ij' yields vertical x-coordinate for image!
        [X, Y, Z] = np.meshgrid(x_interval, y_interval, z_interval,
                                indexing='ij')
        points = np.array([Z.flatten(), Y.flatten(), X.flatten()])

        # Mean/Origin for Gaussian blurring
        origin = np.zeros((self._dimension, 1))

        # Scaling matrix depending on spacing
        S = np.diag(self._spacing)

        # Compute scaled, inverse covariance matrix
        cov_scale_inv = S.dot(np.linalg.inv(cov)).dot(S)

        # Compute Gaussian weights
        values = np.sum((points-origin)*cov_scale_inv.dot(points-origin), 0)
        kernel = np.exp(-0.5*values)
        kernel = kernel/np.sum(kernel)

        # Reshape kernel
        kernel = kernel.reshape(
            x_interval.size, y_interval.size, z_interval.size)

        return kernel

    def get_dx_forward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((1, 1, 2))
        kernel[:] = np.array([1, -1])

        return kernel / self._spacing[0]

    def get_dx_backward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((1, 1, 3))
        kernel[:] = np.array([0, 1, -1])

        return kernel / self._spacing[0]

    def get_dy_forward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((1, 2, 1))
        kernel[:] = np.array([[1], [-1]])

        return kernel / self._spacing[1]

    def get_dy_backward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((1, 3, 1))
        kernel[:] = np.array([[0], [1], [-1]])

        return kernel / self._spacing[1]

    def get_dz_forward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((2, 1, 1))
        kernel[:] = np.array([[[1]], [[-1]]])

        return kernel / self._spacing[2]

    def get_dz_backward_difference(self):

        # kernel = np.zeros((z,y,x))
        kernel = np.zeros((3, 1, 1))
        kernel[:] = np.array([[[0]], [[1]], [[-1]]])

        return kernel / self._spacing[2]



##
# \file solver.py
# \brief      Abstract class to define a numerical solver
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

# Import libraries
import numpy as np
import datetime
from abc import ABCMeta, abstractmethod

import pysitk.python_helper as ph


##
# Abstract class to define a numerical solver
# \date       2017-07-20 23:17:11+0100
#
class Solver(object):
    __metaclass__ = ABCMeta

    ##
    # Store relevant information common for all numerical solvers
    # \date       2017-07-20 23:39:13+0100
    #
    # \param      self     The object
    # \param      x0       Initial value as 1D numpy array
    # \param      x_scale  Characteristic scale of each variable. Setting
    #                      x_scale is equivalent to reformulating the problem
    #                      in scaled variables ``xs = x / x_scale``
    # \param      verbose  Verbose output, bool
    #
    def __init__(self, x0, x_scale, verbose):
        self._x_scale = float(x_scale)
        self._x0 = np.array(x0, dtype=np.float64) / self._x_scale
        self._x = np.array(self._x0)
        self._verbose = verbose
        self._computational_time = datetime.timedelta(seconds=0)
        self._observer = None

    ##
    # Sets the characteristic scale for each variable.
    # \date       2017-08-04 19:09:20+0100
    #
    # \param      self     The object
    # \param      x_scale  Characteristic scale of each variable. Setting
    #                      x_scale is equivalent to reformulating the problem
    #                      in scaled variables ``xs = x / x_scale``
    #
    def set_x_scale(self, x_scale):
        self._x_scale = x_scale

    ##
    # Gets the characteristic scale for each variable.
    # \date       2017-08-04 19:10:13+0100
    #
    # \param      self  The object
    #
    # \return     the characteristic scale for each variable, float or array.
    #
    def get_x_scale(self):
        return self._x_scale

    ##
    # Sets the verbose.
    # \date       2017-08-05 18:26:15+0100
    #
    # \param      self     The object
    # \param      verbose  Turn on/off verbose; bool
    #
    def set_verbose(self, verbose):
        self._verbose = verbose

    ##
    # Gets the verbose.
    # \date       2017-08-05 18:26:44+0100
    #
    # \param      self  The object
    #
    # \return     boolean flag.
    #
    def get_verbose(self):
        return self._verbose

    ##
    # Sets the initial value.
    # \date       2017-09-06 17:32:06+0100
    #
    # \param      self  The object
    # \param      x0    Initial value as 1D numpy array
    #
    def set_x0(self, x0):
        self._x0 = np.array(x0, dtype=np.float64) / self._x_scale
        self._x = np.array(self._x0)

    ##
    # Gets the initial value.
    # \date       2017-08-05 18:30:48+0100
    #
    # \param      self  The object
    #
    # \return     Initial value as 1D numpy array.
    #
    def get_x0(self):
        return np.array(self._x0) * self._x_scale

    ##
    # Gets the obtained numerical estimate to the minimization problem
    # \date       2017-07-20 23:39:36+0100
    #
    # \param      self  The object
    #
    # \return     Numerical solution as 1D numpy array
    #
    def get_x(self):
        return np.array(self._x) * self._x_scale

    ##
    # Gets the computational time it took to obtain the numerical estimate.
    # \date       2017-07-20 23:40:17+0100
    #
    # \param      self  The object
    #
    # \return     The computational time as string
    #
    def get_computational_time(self):
        return self._computational_time

    ##
    # Sets the observer to monitor performance
    # \date       2017-08-05 18:31:18+0100
    #
    # \param      self     The object
    # \param      observer  The observer as Observer object
    #
    def set_observer(self, observer):
        self._observer = observer

    ##
    # Run the numerical solver to obtain the numerical estimate
    # \date       2017-07-20 23:41:08+0100
    #
    # \param      self  The object
    #
    def run(self):

        if self._x0.ndim != 1:
            raise ValueError("Initial value x0 must be a 1D array")

        time_start = ph.start_timing()

        # Execute solver
        self._run()

        # Get computational time
        self._computational_time = ph.stop_timing(time_start)

        if self._verbose:
            ph.print_info("Required computational time: %s" %
                          (self.get_computational_time()))

        if self._observer is not None:
            self._observer.set_computational_time(
                self.get_computational_time())

    @abstractmethod
    def _run(self):
        pass

    @abstractmethod
    def print_statistics(self):
        pass



##
# \file similarity_measures_test.py
#  \brief  Class containing unit tests for lossFunctions
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date Sept 2017

import os
import numpy as np
import unittest
import sys
import matplotlib.pyplot as plt
import pysitk.python_helper as ph

from nsol.similarity_measures import SimilarityMeasures as sim_meas
from nsol.definitions import DIR_TEST


class SimilarityMeasuresTest(unittest.TestCase):

    def setUp(self):
        self.accuracy = 4
        filename = os.path.join(DIR_TEST, "2D_BrainWeb.png")
        self.image = np.array(ph.read_image(filename), dtype=np.float64)
        self.image_scale = self.image * 2
        self.image_off = self.image + 2

        self.x = self.image.flatten()
        self.x_scale = self.image_scale.flatten()
        self.x_off = self.image_off.flatten()

    def test_absolute_errors(self):
        diff = sim_meas.mean_absolute_error(self.x, self.x_off)
        self.assertAlmostEqual(
            diff - np.abs(self.x - self.x_off).mean(), 0, places=self.accuracy)

    def test_squared_errors(self):

        diff = sim_meas.sum_of_squared_differences(self.x, self.x_off)
        self.assertAlmostEqual(
            diff - np.sum(np.square(self.x - self.x_off)), 0,
            places=self.accuracy)

        diff = sim_meas.mean_squared_error(self.x, self.x_off)
        self.assertAlmostEqual(
            diff - np.square(self.x - self.x_off).mean(), 0,
            places=self.accuracy)

        diff = sim_meas.sum_of_squared_differences(self.x, self.x)
        self.assertEqual(np.around(
            diff, decimals=self.accuracy), 0)

        diff = sim_meas.sum_of_squared_differences(self.x, self.x_off)
        error = self.x.size * 4
        self.assertEqual(np.around(
            abs(diff - error), decimals=self.accuracy), 0)

    def test_peak_signal_to_noise_ratio(self):
        diff = sim_meas.peak_signal_to_noise_ratio(self.x, self.x)
        self.assertEqual(np.around(
            diff, decimals=self.accuracy), np.inf)

    def test_normalized_cross_correlation(self):
        diff = sim_meas.normalized_cross_correlation(self.x, self.x)
        self.assertEqual(np.around(
            abs(diff - 1), decimals=self.accuracy), 0)

        diff = sim_meas.normalized_cross_correlation(self.x, -self.x)
        self.assertEqual(np.around(
            abs(diff + 1), decimals=self.accuracy), 0)

        diff = sim_meas.normalized_cross_correlation(self.x, self.x_off)
        self.assertEqual(np.around(
            abs(diff - 1), decimals=self.accuracy), 0)

        diff = sim_meas.normalized_cross_correlation(self.x, self.x_scale)
        self.assertEqual(np.around(
            abs(diff - 1), decimals=self.accuracy), 0)

    def test_dice_score(self):

        # Create naive mask
        x_mask = np.zeros_like(self.x, dtype=bool)
        x_mask[np.where(self.x > 100)] = True

        # ph.show_array(x_mask.reshape(self.image.shape))

        dice = sim_meas.dice_score(x_mask, x_mask)
        self.assertEqual(np.around(
            abs(dice - 1), decimals=self.accuracy), 0)

        dice = sim_meas.dice_score(x_mask, np.zeros_like(x_mask))
        self.assertEqual(np.around(
            dice, decimals=self.accuracy), 0)



#!/usr/bin/python

##
# \file run_denoising_study.py
# \brief      Run TK0L2/TK1L2/TVL2/HuberL2 deconvolution
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import scipy.io
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.linear_operators as LinearOperators
import nsol.data_reader as dr
import nsol.data_writer as dw
import nsol.observer as Observer
import nsol.input_argparser as InputArgparser
import nsol.deconvolution_solver_parameter_study_interface as \
    deconv_interface


def main():

    input_parser = InputArgparser.InputArgparser(
        description="Run TK0L2/TK1L2/TVL2/HuberL2 deconvolution",
    )
    input_parser.add_observation(required=True)
    input_parser.add_reference(required=False)
    input_parser.add_dir_output(required=True)
    input_parser.add_study_name()
    input_parser.add_reconstruction_type(default="TVL2")
    input_parser.add_measures(default=["PSNR", "RMSE", "SSIM", "NCC", "NMI"])
    input_parser.add_blur(default=1)
    input_parser.add_solver(default="PD")
    input_parser.add_iterations(default=50)
    input_parser.add_rho(default=0.1)
    input_parser.add_iter_max(default=10)
    input_parser.add_minimizer(default="lsmr")
    input_parser.add_alpha(default=[0.01])
    input_parser.add_data_loss(default="linear")
    input_parser.add_data_loss_scale(default=1)
    input_parser.add_verbose(default=0)

    # Range for parameter sweeps
    input_parser.add_alpha_range(default=[0.0001, 0.05, 10])
    input_parser.add_data_losses(
        # default=["linear", "arctan"]
    )
    input_parser.add_data_loss_scale_range(
        # default=[0.1, 1.5, 2]
    )

    args = input_parser.parse_args()
    input_parser.print_arguments(args)

    # --------------------------------Read Data--------------------------------
    data_reader = dr.DataReader(args.observation)
    data_reader.read_data()
    observed_nda = data_reader.get_data()

    if args.reference is not None:
        data_reader = dr.DataReader(args.reference)
        data_reader.read_data()
        reference_nda = data_reader.get_data()
        x_ref = reference_nda.flatten()
    else:
        x_ref = None

    if args.blur > 0:
        sigma = np.atleast_1d(args.blur)
        if sigma.ndim != observed_nda.ndim:
            try:
                cov = np.diag(np.ones(observed_nda.ndim)) * sigma**2
            except:
                raise IOError(
                    "Blur information must be either 1- or d-dimensional")

    # ----------------------------Set Up Parameters----------------------------
    parameters = {}
    parameters["alpha"] = np.linspace(
        args.alpha_range[0], args.alpha_range[1], int(args.alpha_range[2]))
    if args.data_losses is not None:
        parameters["data_loss"] = args.data_losses
    if args.data_loss_scale_range is not None:
        parameters["data_loss_scale"] = np.linspace(
            args.data_loss_scale_range[0],
            args.data_loss_scale_range[1],
            int(args.data_loss_scale_range[2]))

    # -------------------------Set Up Parameter Study-------------------------
    dimension = observed_nda.ndim

    b = observed_nda.flatten()
    x0 = observed_nda.flatten()
    x_scale = np.max(observed_nda)

    # Get spacing for blurring operator
    if data_reader.get_image_sitk() is None:
        spacing = np.ones(observed_nda.ndim)
    else:
        spacing = np.array(data_reader.get_image_sitk().GetSpacing())

    # Get linear operators
    linear_operators = eval(
        "LinearOperators.LinearOperators%dD" % dimension)(spacing=spacing)
    A, A_adj = linear_operators.get_gaussian_blurring_operators(cov)
    grad, grad_adj = linear_operators.get_gradient_operators()

    # A: X \rightarrow Y and D: X \rightarrow Z
    X_shape = observed_nda.shape
    Y_shape = A(observed_nda).shape
    Z_shape = grad(observed_nda).shape

    A_1D = lambda x: A(x.reshape(*X_shape)).flatten()
    A_adj_1D = lambda x: A_adj(x.reshape(*Y_shape)).flatten()

    D_1D = lambda x: grad(x.reshape(*X_shape)).flatten()
    D_adj_1D = lambda x: grad_adj(x.reshape(*Z_shape)).flatten()

    if args.study_name is None:
        name = args.reconstruction_type
    else:
        name = args.study_name

    reconstruction_info = {
        "shape": X_shape,
    }

    parameter_study_interface = \
        deconv_interface.DeconvolutionParameterStudyInterface(
            A=A_1D,
            A_adj=A_adj_1D,
            D=D_1D,
            D_adj=D_adj_1D,
            b=b,
            x0=x0,
            alpha=args.alpha[0],
            x_scale=x_scale,
            data_loss=args.data_loss,
            data_loss_scale=args.data_loss_scale,
            iter_max=args.iter_max,
            iterations=args.iterations,
            minimizer=args.minimizer,
            measures=args.measures,
            dimension=dimension,
            reconstruction_type=args.reconstruction_type,
            rho=args.rho,
            dir_output=args.dir_output,
            parameters=parameters,
            name=name,
            reconstruction_info=reconstruction_info,
            x_ref=x_ref,
            tv_solver=args.solver,
            verbose=args.verbose,
        )
    parameter_study_interface.set_up_parameter_study()
    parameter_study = parameter_study_interface.get_parameter_study()

    # Run parameter study
    parameter_study.run()

    print("\nComputational time for Deconvolution Parameter Study %s: %s" %
          (name, parameter_study.get_computational_time()))

    return 0

if __name__ == '__main__':
    main()



##
# \file tikhonov_linear_solver.py
# \brief      Class to define a numerical solver for solving the linear
#             least-squares problems with Tikhonov regularization
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import os
import sys
import scipy
import numpy as np

from nsol.linear_solver import LinearSolver
from nsol.definitions import EPS
from nsol.loss_functions import LossFunctions as lf


##
# Class to estimate the unique minimizer of the convex minimization problem
# max_x [1/2 ||rho( Ax-b )||^2 + alpha/2 ||Bx-b_reg||^2].
# \date       2017-07-20 23:43:19+0100
#
class TikhonovLinearSolver(LinearSolver):

    ##
    # Store relevant information for solving a linear least-squares problem
    # with Tikhonov regularization
    # \date       2017-07-20 23:47:48+0100
    #
    # \param      self             The object
    # \param      A                Function associated to linear operator A:
    #                              X->Y; x -> A(x) with x being a 1D numpy
    #                              array
    # \param      A_adj            Function associated to adjoint linear
    #                              operator A^*: Y -> X; y -> A^*(y)
    # \param      b                Right hand-side of linear system Ax = b as
    #                              1D numpy array
    # \param      B                Function associated to the linear operator
    #                              B: X->Z; x->B(x) with x being a 1D numpy
    #                              array
    # \param      B_adj            Function associated to adjoint linear
    #                              operator B^*: Z->X; z->B^*(z)
    # \param      x0               Initial value as 1D numpy array
    # \param      alpha            Regularization parameter; scalar >= 0
    # \param      b_reg            Right hand-side of linear system associated
    #                              to the regularizer, i.e. Bx = b_reg.
    # \param      data_loss        Data loss function rho specified as string,
    #                              e.g. "linear", "soft_l1", "huber", "cauchy",
    #                              "arctan".
    # \param      data_loss_scale  Value of soft margin between inlier and
    #                              outlier residuals, default is 1.0. The loss
    #                              function is evaluated as rho_(f2) = C**2 *
    #                              rho(f2 / C**2), where C is data_loss_scale.
    #                              This parameter has no effect with
    #                              data_loss='linear', but for other loss
    #                              values it is of crucial importance.
    # \param      minimizer        String defining the used optimizer, i.e.
    #                              "lsmr", "least_squares" or any solver as
    #                              provided by scipy.optimize.minimize
    # \param      iter_max         The iterator maximum
    # \param      x_scale          Characteristic scale of each variable.
    #                              Setting x_scale is equivalent to
    #                              reformulating the problem in scaled
    #                              variables ``xs = x / x_scale``
    # \param      verbose          Verbose output, bool
    # \param      bounds           The bounds
    #
    def __init__(self,
                 A, A_adj,
                 b,
                 B, B_adj,
                 x0,
                 alpha=0.01,
                 b_reg=0,
                 data_loss="linear",
                 data_loss_scale=1,
                 minimizer="lsmr",
                 iter_max=10,
                 x_scale=1,
                 verbose=0,
                 bounds=(0, np.inf)):

        super(self.__class__, self).__init__(
            A=A, A_adj=A_adj, b=b, x0=x0, alpha=alpha, iter_max=iter_max,
            minimizer=minimizer, data_loss=data_loss,
            data_loss_scale=data_loss_scale, x_scale=x_scale, verbose=verbose)

        self._B = B
        self._B_adj = B_adj
        self._b_reg = b_reg / self._x_scale
        self._bounds = bounds

    ##
    # Gets the linear operator B
    # \date       2017-09-06 16:29:11+0100
    #
    # \param      self  The object
    #
    # \return     Forward operator as function: 1D -> 1D
    #
    def get_B(self):
        return self._B

    ##
    # Gets the adjoint linear operator B
    # \date       2017-09-06 16:29:11+0100
    #
    # \param      self  The object
    #
    # \return     Forward operator as function: 1D -> 1D
    #
    def get_B_adj(self):
        return self._B_adj

    def get_b_reg(self):
        return self._b_reg * self._x_scale

    def _run(self):

        if self._minimizer == "lsmr" and self._data_loss != "linear":
            raise ValueError(
                "lsmr solver cannot be used with non-linear data loss")

        elif self._minimizer == "lsq_linear" and self._data_loss != "linear":
            raise ValueError(
                "lsq_linear solver cannot be used with non-linear data loss")

        # Monitor output
        if self._observer is not None:
            self._observer.add_x(self.get_x())

        # Get augmented linear system
        A, b = self._get_augmented_linear_system(self._alpha)

        # Define residual function and its Jacobian
        residual = lambda x: A*x - b
        jacobian_residual = lambda x: A

        # Clip to bounds
        if self._bounds is not None:
            self._x0 = np.clip(self._x0, self._bounds[0], self._bounds[1])

        # Use scipy.sparse.linalg.lsmr
        if self._minimizer == "lsmr" and self._data_loss == "linear":

            # Linear least-squares method
            self._x = scipy.sparse.linalg.lsmr(
                A, b,
                maxiter=self._iter_max,
                show=self._verbose,
                atol=0,
                btol=0)[0]

            if self._bounds is not None:
                # Clip to bounds
                self._x = np.clip(self._x, self._bounds[0], self._bounds[1])

        # Use scipy.optimize.lsq_linear
        elif self._minimizer == "lsq_linear" and self._data_loss == "linear":

            # linear least-squares problem with bounds on the variables
            self._x = scipy.optimize.lsq_linear(
                A, b,
                max_iter=self._iter_max,
                lsq_solver='lsmr',
                lsmr_tol='auto',
                bounds=self._bounds,
                verbose=2*self._verbose,
            ).x

        # Use scipy.optimize.least_squares
        elif self._minimizer == "least_squares":
            # BE AWARE:
            # Loss function is applied to both data and regularization term!
            # Remark: it seems that least_squares solver does not cope with
            # non-linear loss. Maybe because of the use of sparse linear
            # operator?

            # Non-linear least squares algorithm
            self._x = scipy.optimize.least_squares(
                fun=residual,
                jac=jacobian_residual,
                jac_sparsity=jacobian_residual,
                x0=self._x0,
                # method="trf",
                tr_solver='lsmr',
                bounds=self._bounds,
                loss=self._data_loss,
                f_scale=self._data_loss_scale,
                max_nfev=self._iter_max,
                verbose=2*self._verbose,
            ).x

        # Use scipy.optimize.minimize
        else:
            bounds = [[self._bounds[0], self._bounds[1]]] * self._x0.size

            # Define cost function and its Jacobian
            if self._alpha > EPS:
                cost = lambda x: \
                    self._get_cost_data_term(x) + \
                    self._alpha * self._get_cost_regularization_term(x)
                grad_cost = lambda x: \
                    self._get_gradient_cost_data_term(x) + \
                    self._alpha * \
                    self._get_gradient_cost_regularization_term(x)

            else:
                cost = lambda x: self._get_cost_data_term(x)
                grad_cost = lambda x: self._get_gradient_cost_data_term(x)

            self._x = scipy.optimize.minimize(
                method=self._minimizer,
                fun=cost,
                jac=grad_cost,
                x0=self._x0,
                bounds=bounds,
                options={'maxiter': self._iter_max, 'disp': self._verbose}).x

        # Monitor output
        if self._observer is not None:
            self._observer.add_x(self.get_x())

    def _get_augmented_linear_system(self, alpha):

        # With regularization
        if alpha > EPS:

            # Define forward and backward operators
            A_fw = lambda x: self._A_augmented(x, np.sqrt(alpha))
            A_bw = lambda x: self._A_augmented_adj(x, np.sqrt(alpha))

            # Define right-hand side b
            b = np.zeros(A_fw(self._x0).size)
            b[0:self._b.size] = self._b
            b[self._b.size:] = np.sqrt(alpha) * self._b_reg

        # Without regularization
        else:

            # Define forward and backward operators
            A_fw = lambda x: self._A(x)
            A_bw = lambda x: self._A_adj(x)

            # Define right-hand side b
            b = self._b

        # Construct (sparse) linear operator A
        A = scipy.sparse.linalg.LinearOperator(
            shape=(b.size, self._x0.size),
            matvec=A_fw,
            rmatvec=A_bw)

        return A, b

    def _A_augmented(self, x, sqrt_alpha):

        A_augmented_x = np.concatenate((
            self._A(x),
            sqrt_alpha * self._B(x)))

        return A_augmented_x

    def _A_augmented_adj(self, x, sqrt_alpha):

        x_upper = x[:self._b.size]
        x_lower = x[self._b.size:]

        A_augmented_adj_x = self._A_adj(x_upper) + \
            sqrt_alpha * self._B_adj(x_lower)

        return A_augmented_adj_x

    def _get_cost_regularization_term(self, x):
        return 0.5 * np.sum(self._B(x)**2)

    def _get_gradient_cost_regularization_term(self, x):
        return self._B_adj(self._B(x))



##
# \file observer.py
# \brief      Observer class to monitor the iterations of numerical solvers
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import numpy as np

import pysitk.python_helper as ph


##
# Class to monitor the iterations of numerical solvers
# \date       2017-07-23 01:48:54+0100
#
class Observer(object):

    ##
    # Store name of solver to be monitored
    # \date       2017-07-23 00:04:55+0100
    #
    # \param      self  The object
    # \param      name  Name for the solver to monitor, string
    #
    def __init__(self, name="Observer"):
        self._name = name
        self._x_list = []
        self._measures = []
        self._measures_names = []
        self._dic_measures = {}
        self._computational_time = None

    ##
    # Adds a data array to monitor
    # \date       2017-07-23 01:43:49+0100
    #
    # \param      self  The object
    # \param      x     numpy array
    #
    def add_x(self, x):
        self._x_list.append(x)

    ##
    # Sets the name.
    # \date       2017-08-05 02:58:36+0100
    #
    # \param      self  The object
    # \param      name  Name for solver to monitor, string
    #
    def set_name(self, name):
        self._name = name

    ##
    # Clear list of data arrays
    # \date       2017-08-04 19:45:12+0100
    #
    # \param      self  The object
    #
    def clear_x_list(self):
        self._x_list = []

    ##
    # Sets the measures to be computed
    # \date       2017-07-23 01:44:14+0100
    #
    # \param      self          The object
    # \param      measures_dic  Dictionary with measures_dic = { name: f}
    #                           whereby f = lambda x: f(x) the measure to
    #                           evaluate
    #
    def set_measures(self, measures_dic):
        measures_names = list(measures_dic.keys())
        for i in range(0, len(measures_names)):
            self._measures_names.append(measures_names[i])
            self._measures.append(measures_dic[measures_names[i]])
            self._dic_measures.update({measures_names[i]: None})

    ##
    # Gets the measures.
    # \date       2017-08-04 19:51:00+0100
    #
    # \param      self  The object
    #
    # \return     The measures as dictionary.
    #
    def get_measures(self):
        return self._dic_measures

    ##
    # Sets the computational time.
    # \date       2017-07-23 01:45:41+0100
    #
    # \param      self                The object
    # \param      computational_time  The computational time
    #
    def set_computational_time(self, computational_time):
        self._computational_time = computational_time

    ##
    # Calculates the measures.
    # \date       2017-07-23 01:45:50+0100
    #
    # \param      self  The object
    # \post       self._dic_measures contains information on all evaluated
    #             measures
    #
    # \return     The measures.
    #
    def compute_measures(self):
        ph.print_info("Compute measures ... ", newline=False)
        N = len(self._x_list)
        res = np.zeros(N)
        for k, name in enumerate(self._measures_names):
            for i in range(0, N):
                res[i] = self._measures[k](self._x_list[i])
            self._dic_measures[name] = np.array(res)
        print("done")

    ##
    # Gets the list of collected data arrays
    # \date       2017-07-23 01:46:00+0100
    #
    # \return     list of numpy arrays.
    #
    def get_x_list(self):
        return self._x_list

    ##
    # Gets the name.
    # \date       2017-07-23 01:47:03+0100
    #
    # \param      self  The object
    #
    # \return     The name as string
    #
    def get_name(self):
        return self._name

    ##
    # Gets the measures.
    # \date       2017-07-23 01:47:11+0100
    #
    # \param      self  The object
    #
    # \return     The measures as dictionary with format {measure_name: array}.
    #
    def get_measures(self):
        return self._dic_measures

    ##
    # Gets the computational time.
    # \date       2017-07-23 01:48:40+0100
    #
    # \param      self  The object
    #
    # \return     The computational time.
    #
    def get_computational_time(self):
        return self._computational_time



##
# \file similarity_measures.py
# \brief      Collection of similarity (and dissimilarity) functions
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#


import skimage.measure
import numpy as np


class SimilarityMeasures(object):

    ##
    # Compute sum of absolute differences (symmetric)
    # \date       2017-08-04 10:09:05+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     sum of absolute differences as scalar value >= 0
    #
    @staticmethod
    def sum_of_absolute_differences(x, x_ref):
        if x.shape != x_ref.shape:
            raise ValueError("Input data shapes do not match")
        return np.sum(np.abs(x - x_ref))

    ##
    # Compute mean of absolute error (symmetric)
    # \date       2019-02-10 11:29:07+0000
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     mean of absolute error as scalar value >= 0
    #
    @staticmethod
    def mean_absolute_error(x, x_ref):
        mae = SimilarityMeasures.sum_of_absolute_differences(x, x_ref)
        mae /= float(x.size)
        return mae

    ##
    # Compute sum of squared differences (symmetric)
    # \date       2017-08-04 10:09:05+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     sum of squared differences as scalar value >= 0
    #
    @staticmethod
    def sum_of_squared_differences(x, x_ref):
        if x.shape != x_ref.shape:
            raise ValueError("Input data shapes do not match")
        return np.sum(np.square(x - x_ref))

    ##
    # Compute mean of squared error (symmetric)
    # \date       2017-08-04 10:09:46+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     mean of squared error as scalar value >= 0
    #
    @staticmethod
    def mean_squared_error(x, x_ref):
        mse = SimilarityMeasures.sum_of_squared_differences(x, x_ref)
        mse /= float(x.size)
        return mse

    ##
    # Compute root mean square error (symmetric)
    # \date       2017-08-04 10:09:46+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     mean of squared error as scalar value >= 0
    #
    @staticmethod
    def root_mean_square_error(x, x_ref):
        return np.sqrt(SimilarityMeasures.mean_squared_error(x, x_ref))

    ##
    # Compute peak signal to noise ratio (non-symmetric)
    # \date       2017-08-04 10:10:13+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     peak signal to noise ratio as scalar value
    #
    @staticmethod
    def peak_signal_to_noise_ratio(x, x_ref):
        mse = SimilarityMeasures.mean_squared_error(x, x_ref)
        return 10 * np.log10(np.max(x_ref) ** 2 / mse)

    ##
    # Compute normalized cross correlation (symmetric)
    # \date       2017-08-04 10:12:14+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     Normalized cross correlation as scalar value between -1 and 1
    #
    @staticmethod
    def normalized_cross_correlation(x, x_ref):
        if x.shape != x_ref.shape:
            raise ValueError("Input data shapes do not match")

        ncc = np.sum((x - x.mean()) * (x_ref - x_ref.mean()))
        ncc /= float(x.size * x.std(ddof=1) * x_ref.std(ddof=1))

        return ncc

    ##
    # Compute structural similarity (symmetric)
    # \see        Wang, Z. et al., 2004. Image Quality Assessment: From Error
    #             Visibility to Structural Similarity. IEEE Transactions on Image
    #             Processing, 13(4), pp.600-612.
    # \date       2017-08-04 10:16:32+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    #
    # \return     Structural similarity as scalar value between 0 and 1
    #
    @staticmethod
    def structural_similarity(x, x_ref):
        return skimage.measure.compare_ssim(x, x_ref)

    ##
    # Compute Shannon entropy
    #
    # Shannon entropy H(X) = - sum p(x) * ln(p(x))
    # \see        Pluim, J.P.W., Maintz, J.B.A. & Viergever, M.A., 2003.
    #             Mutual-information-based registration of medical images: a
    #             survey. IEEE Transactions on Medical Imaging, 22(8), pp.986-1004.
    # \date       2017-08-04 10:21:02+0100
    #
    # \param      x     numpy data array
    # \param      bins  number of bins for histogram, int
    #
    # \return     Shannon entropy as scalar value \in [0, log_b(n)] (e.g.
    #             Wikipedia)
    #
    @staticmethod
    def shannon_entropy(x, bins=100):

        # histogram is computed over flattened array
        hist, bin_edges = np.histogram(x, bins=bins)

        # Compute probabilities
        prob = hist / float(np.sum(hist))

        entropy = - sum([p * np.log(p) for p in prob.flatten() if p != 0])

        return entropy

    ##
    # Compute joint entropy (symmetric)
    #
    # Joint entropy H(X,Y) = - sum p(x,y) * ln(p(x,y))
    # \see        Pluim, J.P.W., Maintz, J.B.A. & Viergever, M.A., 2003.
    #             Mutual-information-based registration of medical images: a
    #             survey. IEEE Transactions on Medical Imaging, 22(8), pp.986-1004.
    # \date       2017-08-04 10:35:18+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    # \param      bins   number of bins for histogram, sequence or int
    #
    # \return     Joint entropy as scalar value >=0
    #
    @staticmethod
    def joint_entropy(x, x_ref, bins=100):

        hist, x_edges, y_edges = np.histogram2d(
            x.flatten(), x_ref.flatten(), bins=bins)

        # Compute probabilities
        prob = hist / float(np.sum(hist))

        jentropy = - sum([p * np.log(p) for p in prob.flatten() if p != 0])
        return jentropy

    ##
    # Compute mutual information (symmetric)
    #
    # MI(X,Y) = - sum p(x,y) * ln( p(x,y) / (p(x) * p(y)) ) = H(X) + H(Y) - H(X,Y)
    # \see        Pluim, J.P.W., Maintz, J.B.A. & Viergever, M.A., 2003.
    #             Mutual-information-based registration of medical images: a
    #             survey. IEEE Transactions on Medical Imaging, 22(8), pp.986-1004.
    # \see        Skouson, M.B., Quji Guo & Zhi-Pei Liang, 2001. A bound on mutual
    #             information for image registration. IEEE Transactions on Medical
    #             Imaging, 20(8), pp.843-846.
    # \date       2017-08-04 10:40:35+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    # \param      bins   number of bins for histogram, sequence or int
    #
    # \return     Mutual information as scalar value >= 0 with upper bound as in
    #             Skouson2001
    #
    @staticmethod
    def mutual_information(x, x_ref, bins=100):
        mi = SimilarityMeasures.shannon_entropy(x, bins=bins)
        mi += SimilarityMeasures.shannon_entropy(x_ref, bins=bins)
        mi -= SimilarityMeasures.joint_entropy(x, x_ref, bins=bins)
        return mi

    ##
    # Compute mutual information (symmetric)
    #
    # NMI(X,Y) = H(X) + H(Y) / H(X,Y)
    # \see        Pluim, J.P.W., Maintz, J.B.A. & Viergever, M.A., 2003.
    #             Mutual-information-based registration of medical images: a
    #             survey. IEEE Transactions on Medical Imaging, 22(8), pp.986-1004.
    # \date       2017-08-04 10:40:35+0100
    #
    # \param      x      numpy data array
    # \param      x_ref  reference numpy data array
    # \param      bins   number of bins for histogram, sequence or int
    #
    # \return     Normalized mutual information as scalar value >= 0
    #
    @staticmethod
    def normalized_mutual_information(x, x_ref, bins=100):
        nmi = SimilarityMeasures.shannon_entropy(x, bins=bins)
        nmi += SimilarityMeasures.shannon_entropy(x_ref, bins=bins)
        nmi /= SimilarityMeasures.joint_entropy(x, x_ref, bins=bins)
        return nmi

    ##
    # Compute Dice's score (Dice's coefficient).
    #
    # dice(A, B) = 2 * |A \cap B | / (|A| + |B|)
    # \see        Dice, L.R., 1945. Measures of the Amount of Ecologic Association
    #             Between Species. Ecology, 26(3), pp.297-302.
    # \date       2017-08-04 11:11:21+0100
    #
    # \param      x      numpy data array, bool
    # \param      x_ref  reference numpy data array, bool
    #
    # \return Dice score between 0 and 1
    #
    @staticmethod
    def dice_score(x, x_ref):

        if x.dtype is not np.dtype(np.bool) or \
                x_ref.dtype is not np.dtype(np.bool):
            raise ValueError("x and x_ref need to be of type boolean")

        dice = 2 * np.sum(x * x_ref)
        dice /= np.sum(x) + np.sum(x_ref)

        return dice

    # Dictionary for all similarity measures
    similarity_measures = {
        "SSD": sum_of_squared_differences.__func__,
        "MAE": mean_absolute_error.__func__,
        "MSE": mean_squared_error.__func__,
        "RMSE": root_mean_square_error.__func__,
        "PSNR": peak_signal_to_noise_ratio.__func__,
        "SSIM": structural_similarity.__func__,
        "NCC": normalized_cross_correlation.__func__,
        "MI": mutual_information.__func__,
        "NMI": normalized_mutual_information.__func__,
    }

    # Values for each similarity measure to 'define' undefined states
    UNDEF = {
        "SSD": np.NaN,
        "MAE": np.NaN,
        "MSE": np.NaN,
        "RMSE": np.NaN,
        "PSNR": np.NaN,
        "SSIM": np.NaN,
        "NCC": np.NaN,
        "MI": np.NaN,
        "NMI": np.NaN,
    }



#!/usr/bin/python

##
# \file run_denoising.py
# \brief      Run TVL1/TVL2/HuberL1/HuberL2 denoising
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import scipy.io
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.linear_operators as LinearOperators
import nsol.admm_linear_solver as admm
import nsol.primal_dual_solver as pd
import nsol.observer as Observer
import nsol.data_reader as dr
import nsol.data_writer as dw
from nsol.similarity_measures import SimilarityMeasures as \
    SimilarityMeasures
from nsol.proximal_operators import ProximalOperators as prox
from nsol.prior_measures import PriorMeasures as prior_meas
import nsol.input_argparser as InputArgparser


def main():

    input_parser = InputArgparser.InputArgparser(
        description="Run TVL1/TVL2/HuberL1/HuberL2 denoising",
    )
    input_parser.add_observation(required=True)
    input_parser.add_result(required=False)
    input_parser.add_reference(required=False)
    input_parser.add_reconstruction_type(default="TVL2")
    input_parser.add_measures(default=["PSNR", "RMSE", "SSIM", "NCC", "NMI"])
    input_parser.add_iterations(default=50)
    input_parser.add_solver(default="PD")
    input_parser.add_rho(default=0.1)
    input_parser.add_alpha(default=[0.03])
    input_parser.add_dir_output_figures(default=None)
    input_parser.add_verbose(default=0)
    args = input_parser.parse_args()
    input_parser.print_arguments(args)

    observers = [None] * len(args.alpha)
    recons = [None] * len(args.alpha)
    data_nda = []
    data_labels = []

    if len(args.alpha) > 1 and args.result is not None:
        ph.print_warning("Multiple alphas overwrite result")
    elif len(args.alpha) == 1 and args.result is None:
        raise IOError("'--result' must be specified")

    if args.result is None:
        ph.print_warning("No output ('--result') provided")

    # --------------------------------Read Data--------------------------------
    data_reader = dr.DataReader(args.observation)
    data_reader.read_data()
    observed_nda = data_reader.get_data()
    dimension = observed_nda.ndim

    data_nda.append(observed_nda)
    if dimension < 3:
        data_labels.append("observed:\n%s" %
                           os.path.basename(args.observation))
    elif dimension == 3:
        data_labels.append(
            os.path.basename(args.observation).split(".")[0])

    if args.reference is not None:
        data_reader = dr.DataReader(args.reference)
        data_reader.read_data()
        reference_nda = data_reader.get_data()
        data_nda.append(reference_nda)

        if dimension < 3:
            data_labels.append("reference:\n%s" %
                               os.path.basename(args.reference))
        elif dimension == 3:
            data_labels.append(
                os.path.basename(args.reference).split(".")[0])

        x_ref = reference_nda.flatten()

    # ------------------------------Set Up Solver------------------------------
    b = observed_nda.flatten()
    x0 = observed_nda.flatten()
    x_scale = np.max(observed_nda)

    linear_operators = eval(
        "LinearOperators.LinearOperators%dD()" % (dimension))
    grad, grad_adj = linear_operators.get_gradient_operators()

    # A: X \rightarrow Y and D: X \rightarrow Z
    X_shape = observed_nda.shape
    Z_shape = grad(observed_nda).shape
    D_1D = lambda x: grad(x.reshape(*X_shape)).flatten()
    D_adj_1D = lambda x: grad_adj(x.reshape(*Z_shape)).flatten()

    if args.reconstruction_type == "TVL1":
        prox_f = lambda x, tau: prox.prox_ell1_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_tv_conj

    elif args.reconstruction_type == "TVL2":
        prox_f = lambda x, tau: prox.prox_ell2_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_tv_conj

    elif args.reconstruction_type == "HuberL1":
        prox_f = lambda x, tau: prox.prox_ell1_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_huber_conj

    elif args.reconstruction_type == "HuberL2":
        prox_f = lambda x, tau: prox.prox_ell2_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_huber_conj

    else:
        raise ValueError("Denoising type '%s' not known" %
                         args.reconstruction_type)

    for i, alpha in enumerate(args.alpha):
        if dimension < 3:
            title_prefix = args.reconstruction_type + \
                " (" + r"$\alpha=%g$)" % alpha
        elif dimension == 3:
            title_prefix = ("%s_alpha=%g" % (
                args.reconstruction_type, alpha)).replace(".", "p")

        if args.solver == "PD":
            solver = pd.PrimalDualSolver(
                prox_f=prox_f,
                prox_g_conj=prox_g_conj,
                B=D_1D,
                B_conj=D_adj_1D,
                L2=8,
                x0=x0,
                alpha=alpha,
                iterations=args.iterations,
                # alg_type=alg_type,
                x_scale=x_scale,
                verbose=args.verbose,
            )

        # ADMM only for TVL2
        # elif args.solver == "ADMM":
        #     solver = admm.ADMMLinearSolver(
        #         A=lambda x: x.flatten(), A_adj=lambda x: x.flatten(),
        #         b=b,
        #         B=D_1D, B_adj=D_adj_1D,
        #         x0=x0,
        #         alpha=alpha,
        #         rho=args.rho,
        #         iterations=args.iterations,
        #         dimension=dimension,
        #         # iter_max=iter_max,
        #         x_scale=x_scale,
        #         verbose=args.verbose,
        #     )
        else:
            raise ValueError("Solver '%s' not known" % args.solver)

        # ---------------------------Similarity Measures-----------------------
        if args.reference is not None:
            measures_dic = {
                m: lambda x, m=m:
                SimilarityMeasures.similarity_measures[m](x, x_ref)
                for m in args.measures}
            observers[i] = Observer.Observer()
            observers[i].set_measures(measures_dic)
        else:
            observers[i] = None
        solver.set_observer(observers[i])

        # ---------------------------Run Reconstruction------------------------
        solver.run()
        recons[i] = np.array(solver.get_x().reshape(*X_shape))
        data_nda.append(recons[i])
        data_labels.append(title_prefix)

        if args.result is not None:
            data_writer = dw.DataWriter(
                recons[i], args.result, data_reader.get_image_sitk())
            data_writer.write_data()

    # ----------------------------Visualize Results----------------------------
    if args.verbose:
        filename_prefix = args.reconstruction_type

        if dimension == 2:
            ph.show_arrays(
                data_nda,
                title=data_labels,
                fig_number=None,
                cmap="jet",
                use_same_scaling=True,
                directory=args.dir_output_figures,
                filename=args.reconstruction_type+"_comparison.pdf",
                save_figure=0 if args.dir_output_figures is None else 1,
                fontsize=8,
            )
        elif dimension == 3:
            images_sitk = [None] * len(data_nda)
            observation_sitk = data_reader.get_image_sitk()
            for i, nda in enumerate(data_nda):
                images_sitk[i] = sitk.GetImageFromArray(nda)
                images_sitk[i].CopyInformation(observation_sitk)
            sitkh.show_sitk_image(images_sitk, label=data_labels)

        if args.reference is not None:
            linestyles = ["-", ":", "-", "-."] * 10

            for ob in observers:
                ob.compute_measures()

            for k in measures_dic.keys():
                x = []
                y = []
                legend = []
                for i, alpha in enumerate(args.alpha):
                    title = args.reconstruction_type + ": %s" % k
                    y.append(observers[i].get_measures()[k])
                    legend.append(r"$\alpha=%g$" % alpha)
                    x.append(range(0, len(y[-1])))
                ph.show_curves(
                    y=y,
                    x=x,
                    xlabel="iteration",
                    labels=legend,
                    title=title,
                    linestyle=linestyles,
                    markers=ph.MARKERS,
                    markevery=10,
                    directory=args.dir_output_figures,
                    filename=filename_prefix+"_"+k+".pdf",
                    save_figure=0 if args.dir_output_figures is None else 1,
                )

    return 0

if __name__ == '__main__':
    main()



##
# \file solvers_test.py
#  \brief  Class containing unit tests for numerical solvers
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date May 2017


# Import libraries
import os
import numpy as np
import unittest
import sys
import matplotlib.pyplot as plt
import pysitk.python_helper as ph

# Import modules
import nsol.linear_operators as LinearOperators
import nsol.noise as Noise
import nsol.tikhonov_linear_solver as tk
import nsol.admm_linear_solver as admm
import nsol.primal_dual_solver as pd
import nsol.observer as observer
from nsol.proximal_operators import ProximalOperators as prox
from nsol.similarity_measures import SimilarityMeasures as sim_meas

from nsol.definitions import DIR_TEST


class SolversTest(unittest.TestCase):

    def _get_operators(self, x_gt, A, A_adj, grad, grad_adj):

        # A: X \rightarrow Y and grad: X \rightarrow Z
        X_shape = x_gt.shape
        Y_shape = A(x_gt).shape
        Z_shape = grad(x_gt).shape

        I_ = lambda x: x.flatten()
        I_adj_ = lambda x: x.flatten()

        A_ = lambda x: A(x.reshape(*X_shape)).flatten()
        A_adj_ = lambda x: A_adj(x.reshape(*Y_shape)).flatten()

        D_ = lambda x: grad(x.reshape(*X_shape)).flatten()
        D_adj_ = lambda x: grad_adj(x.reshape(*Z_shape)).flatten()

        return I_, I_adj_, A_, A_adj_, D_, D_adj_

    def setUp(self):
        self.accuracy = 7

        # Noise level for corrupting image
        self.noise_level = 0.05

        # Standard deviation for blurring
        sigma2 = 1.5

        self.ssd = lambda x: sim_meas.sum_of_squared_differences(x, x_ref)
        self.mse = lambda x: sim_meas.mean_squared_error(x, x_ref)
        self.rmse = lambda x: sim_meas.root_mean_square_error(x, x_ref)
        self.psnr = lambda x: sim_meas.peak_signal_to_noise_ratio(x, x_ref)
        self.ssim = lambda x: sim_meas.structural_similarity(x, x_ref)
        self.ncc = lambda x: sim_meas.normalized_cross_correlation(x, x_ref)
        self.mi = lambda x: sim_meas.mutual_information(x, x_ref)
        self.nmi = lambda x: sim_meas.normalized_mutual_information(x, x_ref)

        # ---------------------------------1D---------------------------------

        # Ground-truth solution
        self.x_gt_1D = np.ones(50) * 50
        self.x_gt_1D[5] = 10
        self.x_gt_1D[16] = 100
        self.x_gt_1D[23] = 150
        self.x_gt_1D[30] = 20

        linear_operators = LinearOperators.LinearOperators1D()
        self.A_1D, self.A_adj_1D = linear_operators.\
            get_gaussian_blurring_operators(sigma2)
        self.grad_1D, self.grad_adj_1D = linear_operators.\
            get_gradient_operators()

        # ---------------------------------2D---------------------------------

        # filename_2D = os.path.join(DIR_TEST, "2D_Lena_256.png")
        filename_2D = os.path.join(DIR_TEST, "2D_BrainWeb.png")

        # Ground-truth solution
        self.x_gt_2D = ph.read_image(filename_2D).astype(np.float64)

        cov = np.diag(np.ones(2)) * sigma2
        linear_operators = LinearOperators.LinearOperators2D()
        self.A_2D, self.A_adj_2D = linear_operators.\
            get_gaussian_blurring_operators(cov)
        self.grad_2D, self.grad_adj_2D = linear_operators.\
            get_gradient_operators()

    ##
    # Test x_scale setting for solver in the context of deblurring in 1D
    # \date       2017-08-04 16:46:07+0100
    #
    def test_x_scale_1D(self):

        x_scale = self.x_gt_1D.max()

        I, I_adj, A, A_adj, D, D_adj = self._get_operators(
            self.x_gt_1D, self.A_1D, self.A_adj_1D,
            self.grad_1D, self.grad_adj_1D)

        # -----------------------------Scale Data-----------------------------

        x_ = self.x_gt_1D / x_scale

        # Blur
        b_ = A(x_)

        # Add noise
        noise = Noise.Noise(b_, seed=1)
        noise.add_poisson_noise(noise_level=self.noise_level)
        b_ = noise.get_noisy_data()

        b = b_.flatten()
        x0 = np.array(b)

        solver = tk.TikhonovLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=1,
        )
        solver.run()
        recon_tk1 = solver.get_x().reshape(*self.x_gt_1D.shape)

        solver = admm.ADMMLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=1,
            dimension=1,
        )
        solver.run()
        recon_admm1 = solver.get_x().reshape(*self.x_gt_1D.shape)

        solver = pd.PrimalDualSolver(
            prox_f=lambda x, tau: prox.prox_linear_least_squares(
                x=x, tau=tau,
                A=A, A_adj=A_adj, b=b, x0=x0, x_scale=1),
            prox_g_conj=prox.prox_tv_conj,
            B=D,
            B_conj=D_adj,
            L2=8,
            x0=x0,
            x_scale=1,
        )
        solver.run()
        recon_pd1 = solver.get_x().reshape(*self.x_gt_1D.shape)

        # ----------------------------Solver Scale----------------------------

        x_ = self.x_gt_1D

        # Blur
        b_ = A(x_)

        # Add noise
        noise = Noise.Noise(b_, seed=1)
        noise.add_poisson_noise(noise_level=self.noise_level)
        b_ = noise.get_noisy_data()

        b = b_.flatten()
        x0 = np.array(b)

        solver = tk.TikhonovLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=x_scale,
        )
        solver.run()
        recon_tk2 = solver.get_x().reshape(*self.x_gt_1D.shape)

        solver = admm.ADMMLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=x_scale,
            dimension=1,
        )
        solver.run()
        recon_admm2 = solver.get_x().reshape(*self.x_gt_1D.shape)

        solver = pd.PrimalDualSolver(
            prox_f=lambda x, tau: prox.prox_linear_least_squares(
                x=x, tau=tau,
                A=A, A_adj=A_adj, b=b, x0=x0, x_scale=x_scale),
            prox_g_conj=prox.prox_tv_conj,
            B=D,
            B_conj=D_adj,
            L2=8,
            x0=x0,
            x_scale=x_scale,
        )
        solver.run()
        recon_pd2 = solver.get_x().reshape(*self.x_gt_1D.shape)

        # ----------------------------Check results----------------------------
        # Check Tikhonov
        self.assertEqual(np.round(
            np.linalg.norm(recon_tk2 - recon_tk1 * x_scale),
            decimals=self.accuracy), 0)

        # Check ADMM
        self.assertEqual(np.round(
            np.linalg.norm(recon_admm2 - recon_admm1 * x_scale),
            decimals=self.accuracy), 0)

        # Check Primal-Dual
        self.assertEqual(np.round(
            np.linalg.norm(recon_pd2 - recon_pd1 * x_scale),
            decimals=self.accuracy), 0)

    ##
    # Test x_scale setting for solver in the context of deblurring in 2D
    # \date       2017-08-04 16:46:07+0100
    #
    def test_x_scale_2D(self):

        x_scale = self.x_gt_2D.max()

        I, I_adj, A, A_adj, D, D_adj = self._get_operators(
            self.x_gt_2D, self.A_2D, self.A_adj_2D,
            self.grad_2D, self.grad_adj_2D)

        # -----------------------------Scale Data-----------------------------

        x_ = self.x_gt_2D / x_scale

        # Blur
        b_ = A(x_)

        # Add noise
        noise = Noise.Noise(b_, seed=1)
        noise.add_poisson_noise(noise_level=self.noise_level)
        b_ = noise.get_noisy_data()

        b = b_.flatten()
        x0 = np.array(b)

        solver = tk.TikhonovLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=1,
        )
        solver.run()
        recon_tk1 = solver.get_x().reshape(*self.x_gt_2D.shape)

        solver = admm.ADMMLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=1,
            dimension=2,
        )
        solver.run()
        recon_admm1 = solver.get_x().reshape(*self.x_gt_2D.shape)

        solver = pd.PrimalDualSolver(
            prox_f=lambda x, tau: prox.prox_linear_least_squares(
                x=x, tau=tau,
                A=A, A_adj=A_adj, b=b, x0=x0, x_scale=1),
            prox_g_conj=prox.prox_tv_conj,
            B=D,
            B_conj=D_adj,
            L2=8,
            x0=x0,
            x_scale=1,
        )
        solver.run()
        recon_pd1 = solver.get_x().reshape(*self.x_gt_2D.shape)

        # ----------------------------Solver Scale----------------------------

        x_ = self.x_gt_2D

        # Blur
        b_ = A(x_)

        # Add noise
        noise = Noise.Noise(b_, seed=1)
        noise.add_poisson_noise(noise_level=self.noise_level)
        b_ = noise.get_noisy_data()

        b = b_.flatten()
        x0 = np.array(b)

        solver = tk.TikhonovLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=x_scale,
        )
        solver.run()
        recon_tk2 = solver.get_x().reshape(*self.x_gt_2D.shape)

        solver = admm.ADMMLinearSolver(
            A=A, A_adj=A_adj,
            B=D, B_adj=D_adj,
            b=b,
            x0=x0,
            x_scale=x_scale,
            dimension=2,
        )
        solver.run()
        recon_admm2 = solver.get_x().reshape(*self.x_gt_2D.shape)

        solver = pd.PrimalDualSolver(
            prox_f=lambda x, tau: prox.prox_linear_least_squares(
                x=x, tau=tau,
                A=A, A_adj=A_adj, b=b, x0=x0, x_scale=x_scale),
            prox_g_conj=prox.prox_tv_conj,
            B=D,
            B_conj=D_adj,
            L2=8,
            x0=x0,
            x_scale=x_scale,
        )
        solver.run()
        recon_pd2 = solver.get_x().reshape(*self.x_gt_2D.shape)

        # ----------------------------Check results----------------------------
        # Check Tikhonov
        self.assertEqual(np.round(
            np.linalg.norm(recon_tk2 - recon_tk1 * x_scale),
            decimals=self.accuracy), 0)

        # Check ADMM
        self.assertEqual(np.round(
            np.linalg.norm(recon_admm2 - recon_admm1 * x_scale),
            decimals=self.accuracy), 0)

        # Check Primal-Dual
        self.assertEqual(np.round(
            np.linalg.norm(recon_pd2 - recon_pd1 * x_scale),
            decimals=self.accuracy), 0)



##
# \file primal_dual_solver.py
# \brief      Implementation of First-order primal-dual algorithm for convex
#             problems min_x [f(x) + alpha g(Bx)] with B being a continuous
#             linear operator.
#
# First-order primal-dual algorithm for convex problems as introduced in
# Chambolle, A. & Pock, T., 2011. A First-Order Primal-Dual Algorithm for
# Convex Problems with Applications to Imaging. Journal of Mathematical Imaging
# and Vision, 40(1), pp.120-145.
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import numpy as np

from nsol.solver import Solver
import pysitk.python_helper as ph


##
# First-order primal-dual algorithm for convex problems
# min_x [f(x) + alpha g(Bx)] with B being a continuous linear operator.
#
class PrimalDualSolver(Solver):

    ##
    # Store all essential variables
    # \date       2017-07-20 22:30:07+0100
    #
    # \param      self         The object
    # \param      prox_f       Proximal operator of f; prox_f: (x, tau) ->
    #                          prox_f(x, tau) = min_y [1/2 ||x-y||^2 + tau
    #                          f(y)]
    # \param      prox_g_conj  Proximal operator of g' with g' being the
    #                          conjugate of a convex, lower-semicontinuous
    #                          function g; typically g acts as regularizer like
    #                          TV or Huber; prox_g_conj: (x, sigma) ->
    #                          prox_g_conj(x, sigma) = min_y [1/2 ||x-y||^2 +
    #                          sigma g'(y)]
    # \param      B            Function associated to continuous linear
    #                          operator B
    # \param      B_conj       The conjugate of the continuous linear operator
    #                          B
    # \param      L2           Squared operator norm of B, i.e.
    #                          L2 = ||B||^2 = ||\nabla||^2 = ||div||^2;
    #                          In 2D: L2 <= 8/h^2 (Chambolle, Pock, p.13)
    #                          In 3D: L2 <= 16/h^2 (similar proof)
    # \param      x0           Initial value, 1D numpy data array
    # \param      alpha        Regularization parameter alpha > 0.
    # \param      iterations   Number of primal-dual iterations, int
    # \param      x_scale      Characteristic scale of each variable. Setting
    #                          x_scale is equivalent to reformulating the
    #                          problem in scaled variables ``xs = x / x_scale``
    # \param      verbose      Verbose output, bool
    # \param      alg_type     Type of algorithm to dynamically update
    #                          parameters for each iteration
    #
    def __init__(self,
                 prox_f,
                 prox_g_conj,
                 B,
                 B_conj,
                 L2,
                 x0,
                 alpha=0.01,
                 iterations=10,
                 x_scale=1.,
                 verbose=0,
                 alg_type="ALG2",
                 ):

        Solver.__init__(self, x0=x0, verbose=verbose, x_scale=x_scale)

        # proximal operator of f
        self._prox_f = prox_f

        # proximal operator of g'
        self._prox_g_conj = prox_g_conj

        # Continuous linear operator B in regularizer term g(Bx)
        self._B = B

        # Conjugate operator of B, i.e. B'
        self._B_conj = B_conj

        # Squared operator norm of B, i.e. L2 = ||B||^2
        self._L2 = float(L2)

        # Regularization parameter in f(x) + alpha g(Bx)
        self._alpha = float(alpha)

        # Number of primal-dual iterations
        self._iterations = iterations

        self._alg_type = alg_type

        # parameter initialization depend on chosen method
        self._get_initial_tau_sigma = {
            "ALG2_AHMOD": self._get_initial_tau_sigma_alg2_ahmod,
            "ALG2": self._get_initial_tau_sigma_alg2,
            "ALG3": self._get_initial_tau_sigma_alg3,
        }

        # parameter updates depend on chosen method
        self._get_update_theta_tau_sigma = {
            "ALG2_AHMOD": self._get_update_theta_tau_sigma_alg2_ahmod,
            "ALG2": self._get_update_theta_tau_sigma_alg2,
            "ALG3": self._get_update_theta_tau_sigma_alg3,
        }

    ##
    # Sets the regularization parameter alpha.
    # \date       2017-08-04 18:54:24+0100
    #
    # \param      self   The object
    # \param      alpha  Regularization  parameter; scalar
    #
    def set_alpha(self, alpha):
        self._alpha = alpha

    ##
    # Gets the regularization parameter alpha.
    # \date       2017-08-04 18:54:59+0100
    #
    # \param      self  The object
    #
    # \return     scalar
    #
    def get_alpha(self):
        return self._alpha

    ##
    # Sets the squared operator norm of B, i.e. L2 = ||B||^2.
    # \date       2017-08-05 19:08:33+0100
    #
    # \param      self  The object
    # \param      L2    scalar value > 0
    #
    def set_L2(self, L2):
        self._L2 = L2

    ##
    # Gets the squared operator norm of B, i.e. L2 = ||B||^2.
    # \date       2017-08-05 19:09:29+0100
    #
    # \param      self  The object
    #
    # \return     Scalar value > 0
    #
    def get_L2(self):
        return self._L2

    ##
    # Sets the type of algorithm to dynamically update parameters for each
    # iteration.
    # \date       2017-08-05 19:16:11+0100
    #
    # \param      self      The object
    # \param      alg_type  string being either 'ALG2', 'ALG2_AHMOD' or 'ALG3'
    #
    def set_alg_type(self, alg_type):
        self._alg_type = alg_type

    ##
    # Gets the type of algorithm to dynamically update parameters for each
    # iteration.
    # \date       2017-08-05 19:17:26+0100
    #
    # \param      self  The object
    #
    # \return     string.
    #
    def get_alg_type(self):
        return self._alg_type

    ##
    # Sets the number of primal-dual iterations.
    # \date       2017-08-05 18:59:12+0100
    #
    # \param      self        The object
    # \param      iterations  Number of primal-dual iterations, integer value
    #
    def set_iterations(self, iterations):
        self._iterations = iterations

    ##
    # Gets the number of primal-dual iterations.
    # \date       2017-08-05 18:59:39+0100
    #
    # \param      self  The object
    #
    # \return     Number of primal-dual iterations, integer value.
    #
    def get_iterations(self):
        return self._iterations

    ##
    # Prints the statistics of the performed optimization
    # \date       2017-07-21 00:01:10+0100
    #
    # \param      self  The object
    # \param      fmt   Format for printing numerical values
    #
    def print_statistics(self, fmt="%.3e"):
        pass

    ##
    # Execute solver
    # \date       2017-08-05 19:10:29+0100
    #
    # \param      self  The object
    #
    def _run(self):

        # Monitor output
        if self._observer is not None:
            self._observer.add_x(self.get_x())

        # regularization parameter lambda as used in Chambolle2011
        lmbda = 1. / self._alpha

        # Dynamic step sizes for primal and dual variable, see p.127
        tau_n, sigma_n, gamma = self._get_initial_tau_sigma[
            self._alg_type](L2=self._L2, lmbda=lmbda)

        x_n = np.array(self._x0)
        x_mean = np.array(self._x0)
        p_n = 0

        for i in range(0, self._iterations):

            # if self._verbose:
            #     ph.print_title("Primal-Dual iteration %d/%d" %
            #                    (i+1, self._iterations))
            # else:
            ph.print_info("Primal-Dual iteration %d/%d" %
                          (i+1, self._iterations))

            # Update dual variable
            p_n = self._prox_g_conj(
                p_n + sigma_n * self._B(x_mean), sigma_n)

            # Update primal variable
            x_np1 = self._prox_f(x_n - tau_n * self._B_conj(p_n), tau_n*lmbda)

            # Update parameter
            theta_n, tau_n, sigma_n = self._get_update_theta_tau_sigma[
                self._alg_type](self._L2, gamma, tau_n, sigma_n)

            # Update mean variable
            x_mean = x_np1 + theta_n * (x_np1 - x_n)

            # Prepare for next iteration
            self._x = x_np1
            x_n = x_np1

            # Monitor output
            if self._observer is not None:
                self._observer.add_x(self.get_x())

        self._x = x_n

    ##
    # Gets the initial step sizes tau_0, sigma_0 and the Lipschitz parameter
    # gamma according to ALG2 method in Chambolle2011, p.133
    #
    # tau_0 and sigma_0 such that tau_0 * sigma_0 * L^2 = 1
    # \date       2017-07-18 17:57:33+0100
    #
    # \param      self   The object
    # \param      L2     Squared operator norm
    # \param      lmbda  Regularization parameter
    #
    # \return     tau0, sigma0, gamma
    #
    def _get_initial_tau_sigma_alg2(self, L2, lmbda):
        # Initial values according to ALG2 in Chambolle2011
        tau0 = 1. / np.sqrt(L2)
        sigma0 = 1. / (L2 * tau0)
        gamma = 0.35 * lmbda
        return tau0, sigma0, gamma

    ##
    # Gets the update of the variable relaxation parameter
    # \f$\theta_n\in[0,1]\f$ and the dynamic step sizes
    # \f$\tau_n,\,\sigma_n>0\f$ for the primal and dual variable, respectively.
    #
    # Update is performed according to ALG2 in Chambolle2011, p.133. It always
    # holds tau_n * sigma_n * L^2 = 1.
    # \date       2017-07-18 18:16:28+0100
    #
    # \param      self     The object
    # \param      L2       Squared operator norm
    # \param      gamma    Lipschitz parameter
    # \param      tau_n    Dynamic step size for primal variable
    # \param      sigma_n  Dynamic step size for dual variable
    #
    # \return     theta_n, tau_n, sigma_n update
    #
    def _get_update_theta_tau_sigma_alg2(self, L2, gamma, tau_n, sigma_n):
        theta_n = 1. / np.sqrt(1. + 2. * gamma * tau_n)
        tau_n = tau_n * theta_n
        sigma_n = sigma_n / theta_n
        return theta_n, tau_n, sigma_n

    ##
    # Gets the initial step sizes tau_0, sigma_0 and the Lipschitz parameter
    # gamma according to ALG2 method in Chambolle2011, p.136
    #
    # tau_0 and sigma_0 such that tau_0 * sigma_0 * L^2 = 1
    # \date       2017-07-18 17:57:33+0100
    #
    # \param      self   The object
    # \param      L2     Squared operator norm
    # \param      lmbda  Regularization parameter
    #
    # \return     tau0, sigma0, theta
    #
    def _get_initial_tau_sigma_alg3(self, L2, lmbda, huber_alpha=0.05):

        # Initial values according to ALG3 in Chambolle2011
        gamma = lmbda
        delta = huber_alpha
        mu = 2. * np.sqrt(gamma * delta / L2)

        # relaxation parameter in [1/(1+mu), 1]
        theta = 1. / (1. + mu)

        # step size dual variable
        sigma = mu / (2. * delta)

        # step size primal variable
        tau = mu / (2. * gamma)

        return tau, sigma, theta

    ##
    # Gets the update of the variable relaxation parameter
    # \f$\theta_n\in[0,1]\f$ and the dynamic step sizes
    # \f$\tau_n,\,\sigma_n>0\f$ for the primal and dual variable, respectively.
    #
    # Update is performed according to ALG2 in Chambolle2011, p.136. It always
    # holds tau_n * sigma_n * L^2 = 1.
    # \date       2017-07-18 18:16:28+0100
    #
    # \param      self     The object
    # \param      L2       Squared operator norm
    # \param      gamma    Lipschitz parameter
    # \param      tau_n    Dynamic step size for primal variable
    # \param      sigma_n  Dynamic step size for dual variable
    #
    # \return     theta_n, tau_n, sigma_n update
    #
    def _get_update_theta_tau_sigma_alg3(self, L2, gamma, tau_n, sigma_n):
        theta_n = gamma  # gamma is used as place holder for tau
        return theta_n, tau_n, sigma_n

    ##
    # Gets the initial step sizes tau_0, sigma_0 and the Lipschitz parameter
    # gamma according to AHMOD, i.e. Arrow-Hurwicz method, in Chambolle2011,
    # p.133
    #
    # tau_0 and sigma_0 such that tau_0 * sigma_0 * L^2 = 4
    # \date       2017-07-18 17:56:36+0100
    #
    # \param      self   The object
    # \param      L2     Squared operator norm
    # \param      lmbda  Regularization parameter
    #
    # \return     tau0, sigma0, gamma
    #
    def _get_initial_tau_sigma_alg2_ahmod(self, L2, lmbda):
        # Initial values according to AHMOD in Chambolle2011
        tau0 = 0.02
        sigma0 = 4. / (L2 * tau0)
        gamma = 0.35 * lmbda
        return tau0, sigma0, gamma

    ##
    # Gets the update of the variable relaxation parameter
    # \f$\theta_n\in[0,1]\f$ and the dynamic step sizes
    # \f$\tau_n,\,\sigma_n>0\f$ for the primal and dual variable, respectively.
    #
    # Update is performed according to AHMOD, i.e. Arrow-Hurwicz method, in
    # Chambolle2011, p.133. It always holds tau_n * sigma_n * L^2 = 4.
    # \date       2017-07-18 18:16:28+0100
    #
    # \param      self     The object
    # \param      L2       Squared operator norm
    # \param      gamma    Lipschitz parameter
    # \param      tau_n    Dynamic step size for primal variable
    # \param      sigma_n  Dynamic step size for dual variable
    #
    # \return     theta_n, tau_n, sigma_n update
    #
    def _get_update_theta_tau_sigma_alg2_ahmod(self,
                                               L2, gamma, tau_n, sigma_n):
        theta_n = 1. / np.sqrt(1. + 2. * gamma * tau_n)
        tau_n = tau_n * theta_n
        sigma_n = sigma_n / theta_n
        return 0., tau_n, sigma_n



##
# \file optimize.py
# \brief      Class with static functions to wrap the functionality of scipy
#             optimizers
#
# Not used for now. Might be relevant at some point if interfaces from one
# solver to the other shall be wrapped. E.g. convert from residuals
# (least_squares) to cost functions (minimize) etc
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import scipy
import numpy as np

class Optimize(object):

    ##
    # Gets the approximate solution lsmr (Linear least-squares method)
    # \date       2017-07-24 10:24:36+0100
    #
    # \param      self      The object
    # \param      A         {matrix, sparse matrix, ndarray, LinearOperator}
    #                       Matrix A in the linear system.
    # \param      b         array_like, shape (m,) Vector b in the linear
    #                       system.
    # \param      max_iter  Maximum number of iterations, int
    # \param      verbose   verbose output, bool
    # \param      conlim    condition limit, float; `lsmr` terminates if an
    #                       estimate of ``cond(A)`` exceeds `conlim`.
    # \param      atol      Stopping tolerances, float
    # \param      btol      Stopping tolerances, float
    # \param      damp      Damping factor for regularized least-squares, float
    #
    # \return     The approximate solution as 1D numpy array.
    #
    @staticmethod
    def get_approximate_solution_lsmr(A, b,
                                      max_iter=None,
                                      verbose=True,
                                      conlim=1e8,
                                      atol=1e-6,
                                      btol=1e-6,
                                      damp=0.0,
                                      ):

        x = scipy.sparse.linalg.lsmr(
            A, b,
            maxiter=max_iter,
            show=verbose,
            atol=atol,
            btol=btol,
            conlim=conlim,
            damp=damp
        )[0]

        return x

    ##
    # Gets the approximate solution lsq linear (linear least-squares problem
    # with bounds)
    # \date       2017-07-24 10:33:24+0100
    #
    # \param      self        The object
    # \param      A           array_like, sparse matrix of LinearOperator,
    #                         shape (m, n)
    # \param      b           array_like, shape (m,)
    # \param      bounds      2-tuple of array_like
    # \param      method      Method to perform minimization. 'trf' or 'bvls'
    # \param      tol         tolerance, float
    # \param      lsq_solver  {None, 'exact', 'lsmr'}
    # \param      lsmr_tol    Tolerance parameters 'atol' and 'btol' for
    #                         `scipy.sparse.linalg.lsmr`
    # \param      max_iter    Maximum number of iterations before termination,
    #                         int
    # \param      verbose     Verbose output, bool
    #
    # \return     The approximate solution as 1D numpy array of shape (m,).
    #
    @staticmethod
    def get_approximate_solution_lsq_linear(A, b,
                                            bounds=(-np.inf, np.inf),
                                            method="trf",
                                            tol=1e-10,
                                            lsq_solver=None,
                                            lsmr_tol=None,
                                            max_iter=None,
                                            verbose=2,
                                            ):
        res = scipy.optimize.lsq_linear(
            A, b,
            bounds=bounds,
            method=method,
            tol=tol,
            lsq_solver=lsq_solver,
            lsmr_tol=lsmr_tol,
            max_iter=max_iter,
            verbose=verbose
        )

        return res.x

    ##
    # Gets the approximate solution nnls (FORTRAN non-negative least squares
    # solver).
    # \date       2017-07-24 10:48:06+0100
    #
    # \param      self  The object
    # \param      A     numpy array of shape (m, n)
    # \param      b     numpy array of shape (m,)
    #
    # \return     The approximate solution as 1D numpy array of shape (n,).
    #
    @staticmethod
    def get_approximate_solution_nnls(A, b):
        return scipy.optimize.nnls(A, b)

    ##
    # Gets the approximate solution least squares (non-linear least-squares
    # problem with bounds).
    # \date       2017-07-24 10:54:18+0100
    #
    # \param      self          The object
    # \param      fun           Function which computes the vector of
    #                           residuals, callable
    # \param      x0            initial value, array_like
    # \param      jac           {'2-point', '3-point', 'cs', callable}. Method
    #                           of computing the Jacobian matrix (an m-by-n
    #                           matrix, where element (i, j) is the partial
    #                           derivative of f[i] with respect to x[j]),
    # \param      bounds        Lower and upper bounds on independent
    #                           variables, 2-tuple of array_like
    # \param      method        {'trf', 'dogbox', 'lm'}. Algorithm to perform
    #                           minimization
    # \param      ftol          float. Tolerance for termination by the change
    #                           of the cost function
    # \param      xtol          float. Tolerance for termination by the change
    #                           of the independent variables
    # \param      gtol          float. Tolerance for termination by the norm of
    #                           the gradient
    # \param      x_scale       array_like or 'jac'. Characteristic scale of
    #                           each variable
    # \param      loss          str or callable. Determines the loss function.
    #                           {'linear', 'soft_l1', 'huber', 'cauchy',
    #                           'arctan'}
    # \param      f_scale       float. Value of soft margin between inlier and
    #                           outlier residuals
    # \param      diff_step     None or array_like. Determines the relative
    #                           step size for the finite difference
    #                           approximation of the Jacobian
    # \param      tr_solver     {None, 'exact', 'lsmr'}. Method for solving
    #                           trust-region subproblems, relevant only for
    #                           'trf' and 'dogbox' methods
    # \param      tr_options    dict. Keyword options passed to trust-region
    #                           solver
    # \param      jac_sparsity  {None, array_like, sparse matrix}. Defines the
    #                           sparsity structure of the Jacobian matrix for
    #                           finite difference estimation, its shape must be
    #                           (m, n)
    # \param      max_nfev      None or int. Maximum number of function
    #                           evaluations before the termination
    # \param      verbose       Verbose output, bool
    # \param      args          tuple and dict. Additional arguments passed to
    #                           `fun` and `jac`
    # \param      kwargs        tuple and dict. Additional arguments passed to
    #                           `fun` and `jac`
    #
    # \return     The approximate solution least squares as 1D numpy array
    #
    @staticmethod
    def get_approximate_solution_least_squares(fun, x0,
                                               jac='2-point',
                                               bounds=(-np.inf, np.inf),
                                               method='trf',
                                               ftol=1e-8,
                                               xtol=1e-8,
                                               gtol=1e-8,
                                               x_scale=1.0,
                                               loss='linear',
                                               f_scale=1.0,
                                               diff_step=None,
                                               tr_solver=None,
                                               tr_options={},
                                               jac_sparsity=None,
                                               max_nfev=None,
                                               verbose=2,
                                               args=(),
                                               kwargs={}):

        res = scipy.optimize.least_squares(
            fun, x0,
            jac=jac,
            bounds=bounds,
            method=method,
            ftol=ftol,
            xtol=xtol,
            gtol=gtol,
            x_scale=x_scale,
            loss=loss,
            f_scale=f_scale,
            diff_step=diff_step,
            tr_solver=tr_solver,
            tr_options=tr_options,
            jac_sparsity=jac_sparsity,
            max_nfev=max_nfev,
            verbose=verbose,
            args=args,
            kwargs=kwargs,
        )

        return res.x

    ##
    # Gets the approximate solution minimize.
    # \date       2017-07-24 11:28:20+0100
    #
    # \param      self         The object
    # \param      fun          callable. Objective function
    # \param      x0           ndarray. Initial value.
    # \param      args         Tuple. Extra arguments passed to the objective
    #                          function and its derivatives (Jacobian, Hessian)
    # \param      method       str or callable.
    # \param      jac          bool or callable. Jacobian (gradient) of
    #                          objective function
    # \param      hess         callable, optional. Hessian (matrix of
    #                          second-order derivatives) of objective function
    #                          or Hessian of objective function times an
    #                          arbitrary vector p
    # \param      hessp        callable, optional. Hessian (matrix of
    #                          second-order derivatives) of objective function
    #                          or Hessian of objective function times an
    #                          arbitrary vector p
    # \param      bounds       sequence
    # \param      constraints  dict or sequence of dict. Constraints definition
    #                          (only for COBYLA and SLSQP)
    # \param      tol          float. Tolerance for termination
    # \param      callback     callable. Called after each iteration
    # \param      options      dict. A dictionary of solver options
    #
    # \return     The approximate solution minimize.
    #
    @staticmethod
    def get_approximate_solution_minimize(fun, x0,
                                          args=(),
                                          method=None,
                                          jac=None,
                                          hess=None,
                                          hessp=None,
                                          bounds=None,
                                          constraints=(),
                                          tol=None,
                                          callback=None,
                                          options=None,
                                          ):
        res = scipy.optimize.minimize(
            fun=fun,
            x0=x0,
            args=args,
            method=method,
            jac=jac,
            hess=hess,
            hessp=hessp,
            bounds=bounds,
            constraints=constraints,
            tol=tol,
            callback=callback,
            options=options,
        )

        return res.x



##
# \file parameter_study.py
# \brief Abstract class to provide basis for solver specific parameter studies
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Aug 2017
#

import os
import six
import itertools
import numpy as np
import SimpleITK as sitk
import matplotlib.pyplot as plt
from abc import ABCMeta, abstractmethod

import pysitk.python_helper as ph

from nsol.parameter_study import ParameterStudy


##
# Abstract class to run parameter studies for solvers
# \date       2017-08-05 19:19:21+0100
#
class SolverParameterStudy(ParameterStudy):
    __metaclass__ = ABCMeta

    def __init__(self,
                 solver,
                 parameters,
                 observer,
                 dir_output,
                 name,
                 reconstruction_info,
                 ):

        ParameterStudy.__init__(self, directory=dir_output, name=name)

        self._solver = solver
        self._parameters = parameters
        self._observer = observer
        self._reconstruction_info = reconstruction_info

    ##
    # Run parameter study and write results to specified files
    # \date       2017-08-05 19:27:16+0100
    #
    # \param      self  The object
    #
    def run(self):

        # Reset observer and attach to solver in case this has not happened
        self._observer.set_name(self._name)
        self._observer.clear_x_list()
        self._solver.set_observer(self._observer)

        # Create files where output is written to
        self._create_file_parameters()
        self._create_files_measures()
        self._create_file_computational_time()

        time_start = ph.start_timing()

        self._run()

        # Get computational time in seconds
        self._computational_time = ph.stop_timing(time_start)

    def get_computational_time(self):
        return self._computational_time

    ##
    # Gets the parameters specified for parameter study.
    # \date       2017-08-05 23:35:09+0100
    #
    # \param      self  The object
    #
    # \return     The parameters as dictionary.
    #
    def get_parameters(self):
        return self._parameters

    def _run(self):

        dic_parameter = {}
        dic_x = {k: v for k, v in six.iteritems(self._reconstruction_info)}

        # Create list from itertools. Then total number of iterations known
        iterations = list(itertools.product(*self._parameters.values()))

        for i, vals in enumerate(iterations):

            ph.print_title("%s: Iteration %d/%d" %
                           (self._name, i+1, len(iterations)))

            for j, key in enumerate(self._parameters.keys()):

                # Update varying parameters of solver
                map(eval("self._solver.set_" + key), [vals[j]])

                # Store current parameter values and print it on the screen
                dic_parameter[key] = eval(
                    "str(self._solver.get_" + key + "())")
                ph.print_info(key + " = %s" % (dic_parameter[key]))

            # Execute solver
            self._solver.run()

            # Compute similarity measures for solver estimate iterations
            self._observer.compute_measures()

            # Write all measure results to file for all iterations
            measures = self._observer.get_measures()
            for measure in measures:

                # Write measure results as line to the measure's file
                nda = measures[measure].reshape(1, -1)
                self._add_to_file_measures(measure, nda)

            # Write required computational time
            self._add_to_file_computational_time(
                self._observer.get_computational_time())

            # Write current parameter values to file
            self._add_to_file_parameters(dic_parameter)

            # Write last iteration of reconstruction to file
            # Data array is associated to line in parameters file
            var = str(i)
            dic_x[var] = np.array(self._observer.get_x_list()[-1])

            # Write results of all obtained arrays at each iteration
            # (Previous results will be overwritten at each iteration)
            self._write_to_file_reconstructions(dic_x)

            # Clear observer for next parameter selection
            self._observer.clear_x_list()

            # Reset solver to initial value
            self._solver.set_x0(self._solver.get_x0())

    ##
    # Creates file where all parameters configurations are stored.
    # \date       2017-08-05 19:25:04+0100
    #
    # \param      self  The object
    #
    def _create_file_parameters(self):

        # Build header
        header = self._get_fileheader()
        header += "## " + ("\t").join(self._parameters.keys()) + "\n"

        # Write file
        ph.write_to_file(self._get_path_to_file_parameters(), header, "w")

    ##
    # Creates a file for each chosen measure where all solver iterations are
    # stored for each parameter configuration
    # \date       2017-08-05 19:25:28+0100
    #
    # \param      self  The object
    #
    def _create_files_measures(self):

        measures = self._observer.get_measures()
        for measure in measures.keys():

            # Build header
            header = self._get_fileheader()
            header += "## " + measure + " for iteration 0 to n\n"

            # Write file
            ph.write_to_file(self._get_path_to_file_measures(measure),
                             header, "w")

    ##
    # Creates a file to computational time required to run each parameter
    # configuration.
    # \date       2017-08-05 19:26:09+0100
    #
    # \param      self  The object
    #
    def _create_file_computational_time(self):

        # Build header
        header = self._get_fileheader()
        header += "## " + "Computational time measured for n iterations" + "\n"

        # Write file
        ph.write_to_file(self._get_path_to_file_computational_time(),
                         header, "w")

    ##
    # Adds parameter configuration to parameter file.
    # \date       2017-08-04 22:06:58+0100
    #
    # \param      self            The object
    # \param      dic_parameters  dictionary holding the parameters and their
    #                             values
    #
    def _add_to_file_parameters(self, dic_parameters):
        text = ("\t").join(dic_parameters.values())
        text += "\n"
        ph.write_to_file(self._get_path_to_file_parameters(), text, "a")

    ##
    # Adds obtained results for all iterations to the measure's file
    # \date       2017-08-05 19:19:51+0100
    #
    # \param      self     The object
    # \param      measure  ID of measure which was evaluated
    # \param      nda      Numpy data array as (1 x N) data array
    #
    def _add_to_file_measures(self, measure, nda):
        ph.write_array_to_file(self._get_path_to_file_measures(measure), nda)

    ##
    # Adds measured computational time to file.
    # \date       2017-08-05 19:23:35+0100
    #
    # \param      self                The object
    # \param      computational_time  computational time as timedelta object
    #
    def _add_to_file_computational_time(self, computational_time):
        text = str(computational_time)
        text += "\n"
        ph.write_to_file(
            self._get_path_to_file_computational_time(), text, "a")

    ##
    # Adds obtained results for all iterations to the reconstruction's file
    # \date       2017-08-05 19:19:51+0100
    #
    # \param      self     The object
    # \param      measure  ID of measure which was evaluated
    # \param      nda      Numpy data array as (1 x N) data array
    #
    def _write_to_file_reconstructions(self, dic):
        np.savez_compressed(self._get_path_to_file_reconstructions(), **dic)
        ph.print_info("File '%s' written" %
                      (self._get_path_to_file_reconstructions()))

    ##
    # Get solver-specific header for all files to be written
    # \date       2017-08-05 19:24:23+0100
    #
    # \param      self  The object
    #
    # \return     The fileheader as tring.
    #
    @abstractmethod
    def _get_fileheader(self):
        pass



##
# \file InterfaceDeconvolutionParameterStudy.py
# \brief      Common interface to run deconvolution reconstruction and
#             parameter study
#
# Idea is to provide a simple interface so that the solver/parameter study for
# the deconvolution problem ||Ax - b|| + alpha g(x) can be performed in any
# additional project outside nsol for the reconstruction methods:
#   -# TK0L2: least-squares optimization with zeroth-order Tikhonov reg.
#   -# TK1L2: least-squares optimization with first-order Tikhonov reg.
#   -# TVL2: least-squares optimization with isotropic total variation reg.
#   -# HuberL2: least-squares optimization with Huber norm reg.
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import scipy.io
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.tikhonov_linear_solver as tk
import nsol.admm_linear_solver as admm
import nsol.primal_dual_solver as pd
import nsol.observer as Observer
import nsol.tikhonov_linear_solver_parameter_study as tkparam
import nsol.primal_dual_solver_parameter_study as pdparam
import nsol.admm_linear_solver_parameter_study as admmparam
from nsol.similarity_measures import SimilarityMeasures as \
    SimilarityMeasures
from nsol.proximal_operators import ProximalOperators as prox
from nsol.prior_measures import PriorMeasures as prior_meas
from nsol.loss_functions import LossFunctions as loss_fun


##
# Interface to get desired TK0L2/TK1L2/TVL2/HuberL2 solver and associated
# (similarity) measures for facilitated analysis.
# \date       2017-09-13 21:24:08+0100
#
class DeconvolutionSolverStudyInterface(object):

    ##
    # Store parameters to get solver.
    #
    # Note that not all variables are used for all solvers! E.g. data_loss, rho
    # \date       2017-09-13 21:29:01+0100
    #
    # \param      self                 The object
    # \param      A                    Function associated with linear operator
    #                                  A: X->Y; x -> A(x) with x being a 1D
    #                                  numpy array
    # \param      A_adj                Function associated with adjoint linear
    #                                  operator A^*: Y -> X; y -> A^*(y)
    # \param      D                    Function associated with the
    #                                  differential/gradient operator D: X->Z;
    #                                  x->D(x) with x being a 1D numpy array
    # \param      D_adj                Function associated with adjoint linear
    #                                  operator B^*: Z->X; z->B^*(z)
    # \param      b                    Right hand-side of linear system Ax = b
    #                                  as 1D numpy array
    # \param      x0                   Initial value as 1D numpy array
    # \param      alpha                The alpha
    # \param      x_scale              Characteristic scale of each variable.
    #                                  Setting x_scale is equivalent to
    #                                  reformulating the problem in scaled
    #                                  variables ``xs = x / x_scale``
    # \param      iter_max             The iterator maximum
    # \param      iterations           The iterations
    # \param      minimizer            String defining the used optimizer, i.e.
    #                                  "lsmr", "least_squares" or any solver as
    #                                  provided by scipy.optimize.minimize
    # \param      measures             The measures
    # \param      reconstruction_type  The reconstruction type
    # \param      dimension            The dimension
    # \param      rho                  regularization parameter of augmented
    #                                  Lagrangian term; scalar > 0
    # \param      x_ref                Reference data array; required for the
    #                                  evaluation of (optional) similarity
    #                                  measures
    # \param      data_loss            Data loss function rho specified as
    #                                  string, e.g. "linear", "soft_l1",
    #                                  "huber", "cauchy", "arctan".
    # \param      data_loss_scale      Value of soft margin between inlier and
    #                                  outlier residuals, default is 1.0. The
    #                                  loss function is evaluated as rho_(f2) =
    #                                  C**2 * rho(f2 / C**2), where C is
    #                                  data_loss_scale. This parameter has no
    #                                  effect with data_loss='linear', but for
    #                                  other loss values it is of crucial
    #                                  importance.
    # \param      tv_solver            Decide on the used TV solver. Either
    #                                  "PD" (Primal-Dual) or "ADMM"
    # \param      verbose              Verbose output, bool
    #
    def __init__(self,
                 A,
                 A_adj,
                 D,
                 D_adj,
                 b,
                 x0,
                 alpha,
                 x_scale,
                 iter_max,
                 iterations,
                 minimizer,
                 measures,
                 reconstruction_type,
                 dimension,
                 L2=8,
                 rho=0.5,
                 x_ref=None,
                 x_ref_mask=None,
                 data_loss="linear",
                 data_loss_scale=1,
                 tv_solver="PD",
                 verbose=0,
                 ):

        self._A = A
        self._A_adj = A_adj
        self._D = D
        self._D_adj = D_adj
        self._b = b
        self._x0 = x0
        self._alpha = alpha
        self._data_loss = data_loss
        self._data_loss_scale = data_loss_scale
        self._x_scale = x_scale
        self._iter_max = iter_max
        self._iterations = iterations
        self._minimizer = minimizer
        self._measures = measures
        self._reconstruction_type = reconstruction_type
        self._x_ref = x_ref
        self._x_ref_mask = x_ref_mask
        self._dimension = dimension
        self._tv_solver = tv_solver
        self._L2 = L2
        self._rho = rho
        self._verbose = verbose

        self._set_up_solver = {
            "TK0L2": self._set_up_solver_TK0L2,
            "TK1L2": self._set_up_solver_TK1L2,
            "TVL2": self._set_up_solver_TVL2,
            "HuberL2": self._set_up_solver_HuberL2,
        }

        self._append_reg_and_data_costs = {
            "TK0L2": self._append_reg_and_data_costs_TK0L2,
            "TK1L2": self._append_reg_and_data_costs_TK1L2,
            "TVL2": self._append_reg_and_data_costs_TVL2,
            "HuberL2": self._append_reg_and_data_costs_HuberL2,
        }

    def set_up_solver(self):
        self._solver = self._set_up_solver[self._reconstruction_type]()

    def set_up_measures(self):

        # Add all selected measures and append reg./prior and data costs
        if self._x_ref is not None:

            if not isinstance(self._x_ref, np.ndarray):
                raise ValueError("Reference x_ref must be of type 1D np.array")

            if self._x_ref.shape != self._x0.shape:
                raise ValueError(
                    "Initial value x0 and reference x_ref arrays "
                    "must be of same shape")

            # Reduce evaluation to mask
            if self._x_ref_mask is not None:
                if self._x_ref.shape != self._x_ref_mask.shape:
                    raise ValueError(
                        "Reference x_ref and reference mask x_ref_mask arrays "
                        "must be of same shape")
                indices = np.where(self._x_ref_mask > 0)

            # Evaluate on entire array
            else:
                indices = np.where(self._x_ref != np.inf)

            measures_dic = {
                m: lambda x, m=m:
                SimilarityMeasures.similarity_measures[m](
                    x[indices], self._x_ref[indices])
                for m in self._measures}
        else:
            measures_dic = {}
        self._append_reg_and_data_costs[self._reconstruction_type](
            measures_dic)

        self._measures_dic = measures_dic

    def get_solver(self):
        if self._solver is None:
            raise RuntimeError("Run 'set_up_solver' first")
        return self._solver

    def get_measures(self):
        if self._measures_dic is None:
            raise RuntimeError("Run 'set_up_measures' first")
        return self._measures_dic

    def _set_up_solver_TK0L2(self):

        solver = tk.TikhonovLinearSolver(
            A=self._A,
            A_adj=self._A_adj,
            B=lambda x: x.flatten(),
            B_adj=lambda x: x.flatten(),
            b=self._b,
            alpha=self._alpha,
            x0=self._x0,
            x_scale=self._x_scale,
            data_loss=self._data_loss,
            data_loss_scale=self._data_loss_scale,
            iter_max=self._iter_max,
            minimizer=self._minimizer,
            verbose=self._verbose,
        )
        return solver

    def _set_up_solver_TK1L2(self):

        solver = tk.TikhonovLinearSolver(
            A=self._A,
            A_adj=self._A_adj,
            B=self._D,
            B_adj=self._D_adj,
            b=self._b,
            alpha=self._alpha,
            x0=self._x0,
            x_scale=self._x_scale,
            data_loss=self._data_loss,
            data_loss_scale=self._data_loss_scale,
            iter_max=self._iter_max,
            minimizer=self._minimizer,
            verbose=self._verbose,
        )
        return solver

    def _set_up_solver_TVL2(self):

        if self._tv_solver == "PD":
            prox_f = lambda x, tau: prox.prox_linear_least_squares(
                x=x, tau=tau,
                A=self._A,
                A_adj=self._A_adj,
                b=self._b,
                x0=self._x0,
                iter_max=self._iter_max,
                data_loss=self._data_loss,
                data_loss_scale=self._data_loss_scale,
                x_scale=self._x_scale)
            solver = pd.PrimalDualSolver(
                prox_f=prox_f,
                prox_g_conj=prox.prox_tv_conj,
                B=self._D,
                B_conj=self._D_adj,
                L2=self._L2,
                alpha=self._alpha,
                x0=self._x0,
                iterations=self._iterations,
                # alg_type=alg_type,
                x_scale=self._x_scale,
                verbose=self._verbose,
            )

        elif self._tv_solver == "ADMM":
            solver = admm.ADMMLinearSolver(
                A=self._A,
                A_adj=self._A_adj,
                b=self._b,
                B=self._D,
                B_adj=self._D_adj,
                alpha=self._alpha,
                x0=self._x0,
                x_scale=self._x_scale,
                data_loss=self._data_loss,
                data_loss_scale=self._data_loss_scale,
                rho=self._rho,
                iterations=self._iterations,
                dimension=self._dimension,
                iter_max=self._iter_max,
                verbose=self._verbose,
            )

        return solver

    def _set_up_solver_HuberL2(self):
        prox_f = lambda x, tau: prox.prox_linear_least_squares(
            x=x, tau=tau,
            A=self._A,
            A_adj=self._A_adj,
            b=self._b,
            x0=self._x0,
            iter_max=self._iter_max,
            x_scale=self._x_scale)
        solver = pd.PrimalDualSolver(
            prox_f=prox_f,
            prox_g_conj=prox.prox_huber_conj,
            B=self._D,
            B_conj=self._D_adj,
            L2=self._L2,
            alpha=self._alpha,
            x0=self._x0,
            iterations=self._iterations,
            # alg_type=alg_type,
            x_scale=self._x_scale,
            verbose=self._verbose,
        )
        return solver

    def _append_reg_and_data_costs_TK0L2(self, measures_dic):
        measures_dic["Reg"] = \
            lambda x: prior_meas.zeroth_order_tikhonov(x)
        measures_dic["Data"] = \
            lambda x: loss_fun.get_ell2_cost_from_residual(
                self._A(x)-self._b,
                loss=self._data_loss,
                f_scale=self._data_loss_scale)

    def _append_reg_and_data_costs_TK1L2(self, measures_dic):
        measures_dic["Reg"] = \
            lambda x: prior_meas.first_order_tikhonov(x, self._D)
        measures_dic["Data"] = \
            lambda x: loss_fun.get_ell2_cost_from_residual(
                self._A(x)-self._b,
                loss=self._data_loss,
                f_scale=self._data_loss_scale)

    def _append_reg_and_data_costs_TVL2(self, measures_dic):
        measures_dic["Reg"] = \
            lambda x: prior_meas.total_variation(x, self._D, self._dimension)
        measures_dic["Data"] = \
            lambda x: loss_fun.get_ell2_cost_from_residual(
                self._A(x)-self._b,
                loss=self._data_loss,
                f_scale=self._data_loss_scale)

    def _append_reg_and_data_costs_HuberL2(self, measures_dic):
        measures_dic["Reg"] = \
            lambda x: prior_meas.huber(x, self._D, self._dimension)
        measures_dic["Data"] = \
            lambda x: loss_fun.get_ell2_cost_from_residual(
                self._A(x)-self._b,
                loss=self._data_loss,
                f_scale=self._data_loss_scale)


class DeconvolutionParameterStudyInterface(DeconvolutionSolverStudyInterface):

    ##
    # Store parameters to get solver
    # \date       2017-09-13 21:29:01+0100
    #
    # \param      self                 The object
    # \param      A                    Function associated with linear operator
    #                                  A: X->Y; x -> A(x) with x being a 1D
    #                                  numpy array
    # \param      A_adj                Function associated with adjoint linear
    #                                  operator A^*: Y -> X; y -> A^*(y)
    # \param      D                    Function associated with the
    #                                  differential/gradient operator D: X->Z;
    #                                  x->D(x) with x being a 1D numpy array
    # \param      D_adj                Function associated with adjoint linear
    #                                  operator B^*: Z->X; z->B^*(z)
    # \param      b                    Right hand-side of linear system Ax = b
    #                                  as 1D numpy array
    # \param      x0                   Initial value as 1D numpy array
    # \param      alpha                The alpha
    # \param      x_scale              Characteristic scale of each variable.
    #                                  Setting x_scale is equivalent to
    #                                  reformulating the problem in scaled
    #                                  variables ``xs = x / x_scale``
    # \param      iter_max             The iterator maximum
    # \param      iterations           The iterations
    # \param      minimizer            String defining the used optimizer, i.e.
    #                                  "lsmr", "least_squares" or any solver as
    #                                  provided by scipy.optimize.minimize
    # \param      measures             The measures
    # \param      dimension            The dimension
    # \param      reconstruction_type  The reconstruction type
    # \param      dir_output           The dir output
    # \param      parameters           The parameters
    # \param      name                 The name
    # \param      reconstruction_info  The reconstruction information
    # \param      rho                  regularization parameter of augmented
    #                                  Lagrangian term; scalar > 0
    # \param      x_ref                Reference data array; required for the
    #                                  evaluation of (optional) similarity
    #                                  measures
    # \param      data_loss            Data loss function rho specified as
    #                                  string, e.g. "linear", "soft_l1",
    #                                  "huber", "cauchy", "arctan".
    # \param      data_loss_scale      Value of soft margin between inlier and
    #                                  outlier residuals, default is 1.0. The
    #                                  loss function is evaluated as rho_(f2) =
    #                                  C**2 * rho(f2 / C**2), where C is
    #                                  data_loss_scale. This parameter has no
    #                                  effect with data_loss='linear', but for
    #                                  other loss values it is of crucial
    #                                  importance.
    # \param      tv_solver            Decide on the used TV solver. Either
    #                                  "PD" (Primal-Dual) or "ADMM"
    # \param      verbose              Verbose output, bool
    #
    def __init__(self,
                 A,
                 A_adj,
                 D,
                 D_adj,
                 b,
                 x0,
                 alpha,
                 x_scale,
                 iter_max,
                 iterations,
                 minimizer,
                 measures,
                 dimension,
                 reconstruction_type,
                 dir_output,
                 parameters,
                 name,
                 reconstruction_info,
                 L2=8,
                 rho=0.5,
                 x_ref=None,
                 x_ref_mask=None,
                 data_loss="linear",
                 data_loss_scale=1,
                 tv_solver="PD",
                 verbose=0,
                 ):

        super(self.__class__, self).__init__(
            A=A,
            A_adj=A_adj,
            D=D,
            D_adj=D_adj,
            b=b,
            x0=x0,
            alpha=alpha,
            data_loss=data_loss,
            data_loss_scale=data_loss_scale,
            x_scale=x_scale,
            iter_max=iter_max,
            iterations=iterations,
            minimizer=minimizer,
            measures=measures,
            reconstruction_type=reconstruction_type,
            L2=L2,
            rho=rho,
            x_ref=x_ref,
            x_ref_mask=x_ref_mask,
            dimension=dimension,
            tv_solver=tv_solver,
            verbose=verbose,
        )
        self._name = name
        self._parameters = parameters
        self._reconstruction_info = reconstruction_info
        self._dir_output = dir_output

        self._solver = None
        self._parameter_study = None

        self._set_up_parameter_study = {
            "TK0L2": self._set_up_parameter_study_TKL2,
            "TK1L2": self._set_up_parameter_study_TKL2,
            "TVL2": self._set_up_parameter_study_TVL2,
            "HuberL2": self._set_up_parameter_study_HuberL2,
        }

    def set_up_parameter_study(self):
        self.set_up_solver()
        self.set_up_measures()

        observer = Observer.Observer()
        observer.set_measures(self._measures_dic)

        self._parameter_study = self._set_up_parameter_study[self._reconstruction_type](
            self._solver, observer,
        )

    def get_parameter_study(self):
        return self._parameter_study

    # def run(self):
    #     self._parameter_study.run()
    #     ph.print_info("Computational Time for Parameter Study: %s" %
    #                   self._parameter_study.get_computational_time())

    def _set_up_parameter_study_TKL2(self, solver, observer):
        parameter_study = tkparam.TikhonovLinearSolverParameterStudy(
            solver, observer,
            dir_output=self._dir_output,
            parameters=self._parameters,
            name=self._name,
            reconstruction_info=self._reconstruction_info,
        )
        return parameter_study

    def _set_up_parameter_study_TVL2(self, solver, observer):

        if self._tv_solver == "PD":
            parameter_study = pdparam.PrimalDualSolverParameterStudy(
                solver, observer,
                dir_output=self._dir_output,
                parameters=self._parameters,
                name=self._name,
                reconstruction_info=self._reconstruction_info,
            )
        elif self._tv_solver == "ADMM":
            parameter_study = admmparam.ADMMLinearSolverParameterStudy(
                solver, observer,
                dir_output=self._dir_output,
                parameters=self._parameters,
                name=self._name,
                reconstruction_info=self._reconstruction_info,
            )
        return parameter_study

    def _set_up_parameter_study_HuberL2(self, solver, observer):
        parameter_study = pdparam.PrimalDualSolverParameterStudy(
            solver, observer,
            dir_output=self._dir_output,
            parameters=self._parameters,
            name=self._name,
            reconstruction_info=self._reconstruction_info,
        )
        return parameter_study



##
# \file loss_functions_test.py
#  \brief  Class containing unit tests for LossFunctions
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date May 2017


# Import libraries
import numpy as np
import unittest
import sys
import pysitk.python_helper as ph

# Import modules
from nsol.loss_functions import LossFunctions as lf


class LossFunctionsTest(unittest.TestCase):

    def setUp(self):

        self.accuracy = 7
        self.m = 500
        self.n = 1000

        self.A = np.random.rand(self.m, self.n)
        self.b = np.random.rand(self.m) * 10
        self.x = np.random.rand(self.n) * 23

    # -----------------------------Loss functions-----------------------------

    def test_linear(self):

        diff = lf.linear(self.b) - self.b

        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        diff = lf.gradient_linear(self.b) - 1
        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

    def test_soft_l1(self):

        soft_l1 = np.zeros_like(self.b)
        soft_l1_grad = np.zeros_like(self.b)

        for i in range(0, self.m):
            e = self.b[i]
            soft_l1[i] = 2*(np.sqrt(1+e)-1)
            soft_l1_grad[i] = 1./np.sqrt(1+e)

        diff = lf.soft_l1(self.b) - soft_l1

        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        diff_grad = lf.gradient_soft_l1(self.b) - soft_l1_grad
        self.assertEqual(np.around(
            np.linalg.norm(diff_grad), decimals=self.accuracy), 0)

    def test_huber(self):

        gamma = 1.3
        gamma2 = gamma * gamma

        huber = np.zeros_like(self.b)
        grad_huber = np.zeros_like(self.b)

        for i in range(0, self.m):
            e = self.b[i]
            if e < gamma2:
                huber[i] = e
                grad_huber[i] = 1
            else:
                huber[i] = 2*gamma*np.sqrt(e) - gamma2
                grad_huber[i] = gamma/np.sqrt(e)

        diff = lf.huber(self.b, gamma=gamma) - huber

        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        diff_grad = lf.gradient_huber(self.b, gamma=gamma) - grad_huber
        self.assertEqual(np.around(
            np.linalg.norm(diff_grad), decimals=self.accuracy), 0)

    def test_cauchy(self):

        cauchy = np.zeros_like(self.b)
        cauchy_grad = np.zeros_like(self.b)

        for i in range(0, self.m):
            e = self.b[i]
            cauchy[i] = np.log(1+e)
            cauchy_grad[i] = 1./(1+e)

        diff = lf.cauchy(self.b) - cauchy

        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        diff_grad = lf.gradient_cauchy(self.b) - cauchy_grad
        self.assertEqual(np.around(
            np.linalg.norm(diff_grad), decimals=self.accuracy), 0)

    def test_arctan(self):

        arctan = np.zeros_like(self.b)
        arctan_grad = np.zeros_like(self.b)

        for i in range(0, self.m):
            e = self.b[i]
            arctan[i] = np.arctan(e)
            arctan_grad[i] = 1. / (1 + e*e)

        diff = lf.arctan(self.b) - arctan

        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        diff_grad = lf.gradient_arctan(self.b) - arctan_grad
        self.assertEqual(np.around(
            np.linalg.norm(diff_grad), decimals=self.accuracy), 0)

    def test_show_curves(self):

        flag_show_curves = 0
        if flag_show_curves:
            M = 50
            steps = 100 * M
            residual = np.linspace(-M, M, steps)

            residual2 = np.array(residual**2)

            losses = []
            grad_losses = []
            jacobians = []
            labels = []

            for loss in ["linear", "soft_l1", "cauchy", "arctan"]:
                for i, f_scale in enumerate([1., 1.5]):
                    if loss == "linear" and i > 0:
                        continue
                    # print loss, f_scale
                    label = loss
                    if loss != "linear":
                        label += "_scale"+str(f_scale)
                    losses.append(np.array(lf.get_loss[loss](
                        f2=residual2, f_scale=f_scale)))
                    grad_losses.append(np.array(lf.get_gradient_loss[loss](
                        f2=residual2, f_scale=f_scale)))
                    jacobians.append(np.array(lf.get_gradient_loss[loss](
                        f2=residual2, f_scale=f_scale)*residual))
                    labels.append(label)

            # losses.append(lf.soft_l1(residual2))
            # grad_losses.append(lf.gradient_soft_l1(residual2))
            # jacobians.append(lf.gradient_soft_l1(residual2)*residual)
            # labels.append("soft_l1")

            for gamma in [1, 1.5]:
                losses.append(np.array(lf.huber(residual2, gamma=gamma)))
                grad_losses.append(
                    np.array(lf.gradient_huber(residual2, gamma=gamma)))
                jacobians.append(np.array(lf.gradient_huber(
                    residual2, gamma=gamma)*residual))
                labels.append("huber(" + str(gamma) + ")")

            x = residual
            ph.show_curves(losses, x=x, labels=labels, xlabel="x",
                           title="Cost rho(x^2)")
            ph.show_curves(grad_losses, x=x, labels=labels, xlabel="x",
                           title="Gradient Loss rho'(x^2)")
            ph.show_curves(jacobians, x=x, labels=labels, xlabel="x",
                           title="Gradient Cost rho'(x^2)*x")

    # --------------------Conversion from residual to cost--------------------

    def test_cost_from_residual_linear(self):

        loss = "linear"

        def f(x):
            nda = np.zeros(3)
            nda[0] = x[0]**2 - 3*x[1]**3 + 5
            nda[1] = 2*x[0] + x[1]**2 - 1
            nda[2] = x[0] + x[1]
            return nda

        def df(x):
            nda = np.zeros((3, 2))
            nda[0, 0] = 2*x[0]
            nda[1, 0] = 2
            nda[2, 0] = 1

            nda[0, 1] = -9*x[1]**2
            nda[1, 1] = 2*x[1]
            nda[2, 1] = 1
            return nda

        X0, X1 = np.meshgrid(np.arange(-5, 5, 0.1), np.arange(0, 10, 0.2))
        points = np.array([X1.flatten(), X0.flatten()])

        cost_gd = lambda x: 0.5 * np.sum(f(x)**2)
        grad_cost_gd = lambda x: np.sum(f(x)[:, np.newaxis]*df(x), 0)

        for i in range(points.shape[1]):
            point = points[:, i]
            diff_cost = cost_gd(point) - \
                lf.get_ell2_cost_from_residual(f(point), loss=loss)
            diff_grad = grad_cost_gd(point) - \
                lf.get_gradient_ell2_cost_from_residual(
                    f(point), df(point), loss=loss)

            self.assertEqual(np.around(
                np.linalg.norm(diff_cost), decimals=self.accuracy), 0)
            self.assertEqual(np.around(
                np.linalg.norm(diff_grad), decimals=self.accuracy), 0)

    def test_linear_least_squares_cost_gradient(self):

        residual = self.A.dot(self.x) - self.b
        ell2 = 0.5*np.sum(residual**2)
        diff = lf.get_ell2_cost_from_residual(f=residual, loss="linear")
        diff -= ell2
        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        grad_ell2 = self.A.transpose().dot(residual)
        diff = lf.get_gradient_ell2_cost_from_residual(
            f=residual,
            jac_f=self.A,
            loss="linear")
        diff -= grad_ell2
        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

    def test_soft_l1_least_squares_cost_gradient(self):

        residual = self.A.dot(self.x) - self.b
        ell2 = 0.5*np.sum(lf.soft_l1(residual**2))
        diff = lf.get_ell2_cost_from_residual(f=residual, loss="soft_l1")
        diff -= ell2
        self.assertEqual(np.around(
            np.linalg.norm(diff), decimals=self.accuracy), 0)

        # Derive analytical cost function; However, unit tests in
        # registration_test.py suggest that all good
    #     grad_ell2 = lf.gradient_soft_l1(res**2).dot(self.A.transpose().dot(residual))
    #     diff = lf.get_gradient_ell2_cost_from_residual(
    #         f=residual,
    #         jac_f=self.A,
    #         loss="linear")
    #     diff -= grad_ell2
    #     self.assertEqual(np.around(
    #         np.linalg.norm(diff), decimals=self.accuracy), 0)



##
# \file input_arparser.py
# \brief      Class holding a collection of possible arguments to parse for
#             scripts
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       September 2017
#

import six
import inspect
import argparse

import pysitk.python_helper as ph

from nsol.similarity_measures import SimilarityMeasures as \
    SimilarityMeasures
from nsol.loss_functions import LossFunctions as \
    LossFunctions
from nsol.definitions import ALLOWED_INPUT_FILE_EXTENSIONS
from nsol.definitions import ALLOWED_NOISE_TYPES

# Allowed input file types
INPUT_FILE_TYPES = "(" + (", ").join(ALLOWED_INPUT_FILE_EXTENSIONS) + ")"
NOISE_TYPES = "(" + (", ").join(ALLOWED_NOISE_TYPES) + ", or none)"

##
# Class holding a collection of possible arguments to parse for scripts
# \date       2017-08-07 01:26:11+0100
#


class InputArgparser(object):

    def __init__(self,
                 description=None,
                 prog=None,
                 epilog="Author: Michael Ebner (michael.ebner.14@ucl.ac.uk)",
                 ):

        kwargs = {}
        if description is not None:
            kwargs['description'] = description
        if prog is not None:
            kwargs['prog'] = prog
        if epilog is not None:
            kwargs['epilog'] = epilog

        self._parser = argparse.ArgumentParser(**kwargs)

    def get_parser(self):
        return self._parser

    def parse_args(self):
        return self._parser.parse_args()

    def print_arguments(self, args, title="Input Parameters:"):
        ph.print_title(title)
        for arg in sorted(vars(args)):
            ph.print_info("%s: " % (arg), newline=False)
            print(getattr(args, arg))

    def add_observation(
        self,
        option_string="--observation",
        type=str,
        help="Path to observation %s." % (INPUT_FILE_TYPES),
        required=True,
    ):
        self._add_argument(dict(locals()))

    def add_filename(
        self,
        option_string="--filename",
        type=str,
        help="Path to filename %s." % (INPUT_FILE_TYPES),
        required=True,
    ):
        self._add_argument(dict(locals()))

    def add_dir_input(
        self,
        option_string="--dir-input",
        type=str,
        help="Input directory.",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_result(
        self,
        option_string="--result",
        type=str,
        help="Specify path for obtained result %s." % (INPUT_FILE_TYPES),
        required=True,
    ):
        self._add_argument(dict(locals()))

    def add_reconstruction_type(
        self,
        option_string="--reconstruction-type",
        type=str,
        help="Define reconstruction type. Allowed values are "
        "'TVL1', 'TVL2', 'HuberL1' and 'HuberL2'.",
        default="TVL1",
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_measures(
        self,
        option_string="--measures",
        type=str,
        nargs="+",
        help="Measures to be evaluated between reference (if given) and "
        "reconstruction %s. " % ("(" + (", ").join(
            SimilarityMeasures.similarity_measures.keys()) + ")"),
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_reference(
        self,
        option_string="--reference",
        type=str,
        help="Path to reference %s. Similarity measures are "
        "computed only if reference is given." % (INPUT_FILE_TYPES),
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_alpha_range(
        self,
        option_string="--alpha-range",
        nargs="+",
        type=float,
        help="Specify regularization parameter array by providing "
        "'First Last Step' information. Array is generated according to "
        "np.linspace(First, Last, Step).",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_data_losses(
        self,
        option_string="--data-losses",
        nargs="+",
        help="Specify data losses to be used %s. " % ("(" + (", ").join(
            LossFunctions.get_loss.keys()) + ")"),
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_data_loss_scale_range(
        self,
        option_string="--data-loss-scale-range",
        nargs="+",
        type=float,
        help="Specify data loss scales by providing 'First Last Step' "
        "information. Array is generated according to "
        "np.linspace(First, Last, Step).",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_study_name(
        self,
        option_string="--study-name",
        type=str,
        help="Name of parameter study without white spaces.",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_dir_output(
        self,
        option_string="--dir-output",
        type=str,
        help="Output directory.",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    def add_dir_output_figures(
        self,
        option_string="--dir-output-figures",
        type=str,
        help="If given, created figures are saved to this directory.",
        default=None,
    ):
        self._add_argument(dict(locals()))

    def add_alpha(
        self,
        option_string="--alpha",
        nargs="+",
        type=float,
        help="Regularization parameter alpha to solve minimization problem "
        "min_x [f(x) + alpha g(x)] with f denoting the data and g the "
        "regularization term, respectively.",
        default=0.03,
    ):
        self._add_argument(dict(locals()))

    def add_blur(
        self,
        option_string="--blur",
        type=float,
        nargs="+",
        help="Specify for Gaussian blurring a single standard deviation "
        "(isotropic blurring) or the standard deviation in each spatial "
        "direction (elliptic blurring)",
        default=0,
    ):
        self._add_argument(dict(locals()))

    def add_noise(
        self,
        option_string="--noise",
        type=str,
        help="Specify type of noise to be applied %s." % NOISE_TYPES,
        default=None,
    ):
        self._add_argument(dict(locals()))

    def add_noise_level(
        self,
        option_string="--noise-level",
        type=float,
        help="Specify noise level to be applied.",
        default=None,
    ):
        self._add_argument(dict(locals()))

    def add_colormap(
        self,
        option_string="--colormap",
        type=str,
        help="Specify colormap type to be applied for visualization "
        "(Only for 2D). "
        "E.g. 'Greys_r', 'jet'",
        default=None,
    ):
        self._add_argument(dict(locals()))

    def add_iter_max(
        self,
        option_string="--iter-max",
        type=int,
        help="Number of maximum iterations for the numerical solver.",
        default=10,
    ):
        self._add_argument(dict(locals()))

    def add_minimizer(
        self,
        option_string="--minimizer",
        type=str,
        help="Choice of minimizer used for the inverse problem associated to "
        "the f(x) + alpha g(x). Possible choices are 'lsmr' or any solver in "
        "scipy.optimize.minimize like 'L-BFGS-B'. Note, in case of a chosen "
        "non-linear data loss only non-linear solvers like 'L-BFGS-B' are "
        "viable.",
        default="lsmr",
    ):
        self._add_argument(dict(locals()))

    def add_rho(
        self,
        option_string="--rho",
        type=float,
        help="Regularization parameter for augmented Lagrangian term for "
        "TV regularization by ADMM.",
        default=0.5,
    ):
        self._add_argument(dict(locals()))

    def add_iterations(
        self,
        option_string="--iterations",
        type=int,
        help="Number of ADMM/Primal Dual iterations.",
        default=10,
    ):
        self._add_argument(dict(locals()))

    def add_solver(
        self,
        option_string="--solver",
        type=str,
        help="Type of solver. Either 'ADMM' or 'PD' to choose between "
        "Alternating Direction Method of Multipliers (ADMM) or "
        "Primal-Dual (PD) approach, respectively.",
        default="PD",
    ):
        self._add_argument(dict(locals()))

    def add_data_loss(
        self,
        option_string="--data-loss",
        type=str,
        help="Loss function rho used for data term, i.e. rho((y_k - A_k x)^2) "
        "Possible choices are 'linear', 'soft_l1, 'huber', 'arctan' and "
        "'cauchy'.",
        default="linear",
    ):
        self._add_argument(dict(locals()))

    def add_data_loss_scale(
        self,
        option_string="--data-loss-scale",
        type=float,
        help="Value of soft margin between inlier and outlier residuals, "
        "default is 1.0. The loss function is evaluated as "
        "rho_(f2) = C**2 * rho(f2 / C**2), where C is data_loss_scale. "
        "This parameter has no effect with data_loss='linear', but for other "
        "loss values it is of crucial importance.",
        default=1,
    ):
        self._add_argument(dict(locals()))

    def add_pd_alg_type(
            self,
            option_string="-pd_alg_type",
            type=str,
            help="Algorithm used to dynamically update parameters for each "
            "iteration of the dual algorithm. "
            "Possible choices are 'ALG2', 'ALG2_AHMOD' and 'ALG3' as "
            "described in Chambolle et al., 2011",
            default="ALG2",
    ):
        self._add_argument(dict(locals()))

    def add_verbose(
        self,
        option_string="--verbose",
        type=int,
        help="Turn on/off verbose output.",
        default=1,
    ):
        self._add_argument(dict(locals()))

    def add_option(
        self,
        option_string="--option",
        nargs=None,
        type=float,
        help="Add option.",
        default=None,
        required=False,
    ):
        self._add_argument(dict(locals()))

    ##
    # Adds an argument to argument parser.
    #
    # Rationale: Make interface as generic as possible so that function call
    # works regardless the name of the desired option
    # \date       2017-08-06 21:54:51+0100
    #
    # \param      self     The object
    # \param      allvars  all variables set at respective function call as
    #                      dictionary
    #
    def _add_argument(self, allvars):

        # Skip variable 'self'
        allvars.pop('self')

        # Get name of argument to add
        option_string = allvars.pop('option_string')

        # Build dictionary for additional, optional parameters
        kwargs = {}
        for key, value in six.iteritems(allvars):
            kwargs[key] = value

        # Add information on default value in case provided
        if 'default' in kwargs.keys():

            txt_default = " [default: %s]" % (str(kwargs['default']))

            # Case where 'required' key is given:
            if 'required' in kwargs.keys():

                # Only add information in case argument is not mandatory to
                # parse
                if kwargs['default'] is not None and not kwargs['required']:
                    kwargs['help'] += txt_default

            # Case where no such field was provided
            else:
                if kwargs['default'] is not None:
                    kwargs['help'] += txt_default

        # Add argument with its options
        self._parser.add_argument(option_string, **kwargs)



#!/usr/bin/python

##
# \file run_denoising_study.py
# \brief      Run parameter study for denoising problem. Result can be
#             visualized by show_parameter_study.py
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import scipy.io
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.linear_operators as LinearOperators
import nsol.admm_linear_solver as admm
import nsol.primal_dual_solver as pd
import nsol.observer as Observer
import nsol.data_reader as dr
import nsol.data_writer as dw
import nsol.primal_dual_solver_parameter_study as pdparam
import nsol.admm_linear_solver_parameter_study as admmparam
from nsol.similarity_measures import SimilarityMeasures as \
    SimilarityMeasures
from nsol.proximal_operators import ProximalOperators as prox
from nsol.prior_measures import PriorMeasures as PriorMeasures
import nsol.input_argparser as InputArgparser


def main():

    input_parser = InputArgparser.InputArgparser(
        description="Run denoising algorithm study",
    )
    input_parser.add_observation(required=True)
    input_parser.add_reference(required=False)
    input_parser.add_dir_output(required=True)
    input_parser.add_study_name()
    input_parser.add_reconstruction_type(default="TVL2")
    input_parser.add_measures(default=["PSNR", "RMSE", "SSIM", "NCC", "NMI"])
    # input_parser.add_solver(default="PD")
    input_parser.add_iterations(default=200)
    input_parser.add_rho(default=0.1)
    input_parser.add_verbose(default=0)

    # Range for parameter sweeps
    # input_parser.add_alpha_range(default=[0.0001, 0.05, 10])  # TVL2
    input_parser.add_alpha_range(default=[0.01, 1.5, 10])  # TVL1
    input_parser.add_data_losses(
        # default=["linear", "arctan"]
    )
    input_parser.add_data_loss_scale_range(
        # default=[0.1, 1.5, 2]
    )

    args = input_parser.parse_args()
    input_parser.print_arguments(args)

    # --------------------------------Read Data--------------------------------
    data_reader = dr.DataReader(args.observation)
    data_reader.read_data()
    observed_nda = data_reader.get_data()

    if args.reference is not None:
        data_reader = dr.DataReader(args.reference)
        data_reader.read_data()
        reference_nda = data_reader.get_data()
        x_ref = reference_nda.flatten()

    # ------------------------------Set Up Solver------------------------------
    dimension = observed_nda.ndim

    b = observed_nda.flatten()
    x0 = observed_nda.flatten()
    x_scale = np.max(observed_nda)
    # x_scale = np.mean(observed_nda)

    linear_operators = eval(
        "LinearOperators.LinearOperators%dD()" % (dimension))
    grad, grad_adj = linear_operators.get_gradient_operators()

    # A: X \rightarrow Y and D: X \rightarrow Z
    X_shape = observed_nda.shape
    Z_shape = grad(observed_nda).shape
    D_1D = lambda x: grad(x.reshape(*X_shape)).flatten()
    D_adj_1D = lambda x: grad_adj(x.reshape(*Z_shape)).flatten()

    if args.reconstruction_type == "TVL1":
        prox_f = lambda x, tau: prox.prox_ell1_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_tv_conj

    elif args.reconstruction_type == "TVL2":
        prox_f = lambda x, tau: prox.prox_ell2_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_tv_conj

    elif args.reconstruction_type == "HuberL1":
        prox_f = lambda x, tau: prox.prox_ell1_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_huber_conj

    elif args.reconstruction_type == "HuberL2":
        prox_f = lambda x, tau: prox.prox_ell2_denoising(
            x, tau, x0=b, x_scale=x_scale)
        prox_g_conj = prox.prox_huber_conj

    else:
        raise ValueError("Denoising '%s' type not known" %
                         args.reconstruction_type)

    solver = pd.PrimalDualSolver(
        prox_f=prox_f,
        prox_g_conj=prox_g_conj,
        B=D_1D,
        B_conj=D_adj_1D,
        L2=8,
        x0=x0,
        iterations=args.iterations,
        # alg_type=alg_type,
        x_scale=x_scale,
        verbose=args.verbose,
    )

    # ---------------------------Similarity Measures---------------------------
    if args.reference is not None:
        measures_dic = {
            m: lambda x, m=m:
            SimilarityMeasures.similarity_measures[m](x, x_ref)
            for m in args.measures}
    else:
        measures_dic = {}

    if args.reconstruction_type == "TVL1":
        measures_dic["Reg"] = \
            lambda x: PriorMeasures.total_variation(x, D_1D, dimension)
        measures_dic["Data"] = \
            lambda x: SimilarityMeasures.sum_of_absolute_differences(x, x0)

    elif args.reconstruction_type == "TVL2":
        measures_dic["Reg"] = \
            lambda x: PriorMeasures.total_variation(x, D_1D, dimension)
        measures_dic["Data"] = \
            lambda x: SimilarityMeasures.sum_of_squared_differences(x, x0)

    elif args.reconstruction_type == "HuberL1":
        measures_dic["Reg"] = \
            lambda x: PriorMeasures.huber(x, D_1D, dimension)
        measures_dic["Data"] = \
            lambda x: SimilarityMeasures.sum_of_absolute_differences(x, x0)

    elif args.reconstruction_type == "HuberL2":
        measures_dic["Reg"] = \
            lambda x: PriorMeasures.huber(x, D_1D, dimension)
        measures_dic["Data"] = \
            lambda x: SimilarityMeasures.sum_of_squared_differences(x, x0)

    observer = Observer.Observer()
    observer.set_measures(measures_dic)
    solver.set_observer(observer)

    # ----------------------------Set Up Parameters----------------------------
    parameters = {}
    parameters["alpha"] = np.linspace(
        args.alpha_range[0], args.alpha_range[1], int(args.alpha_range[2]))
    if args.data_losses is not None:
        parameters["data_loss"] = args.data_losses
    if args.data_loss_scale_range is not None:
        parameters["data_loss_scale"] = np.linspace(
            args.data_loss_scale_range[0],
            args.data_loss_scale_range[1],
            int(args.data_loss_scale_range[2]))

    # -------------------------Set Up Parameter Study-------------------------
    if args.study_name is None:
        name = args.reconstruction_type
    else:
        name = args.study_name

    parameter_study = pdparam.PrimalDualSolverParameterStudy(
        solver, observer,
        dir_output=args.dir_output,
        parameters=parameters,
        name=name,
        reconstruction_info={
            "shape": X_shape,
        }
    )

    # Run parameter study
    parameter_study.run()

    print("\nComputational time for Denoising Parameter Study %s: %s" %
          (name, parameter_study.get_computational_time()))

    return 0

if __name__ == '__main__':
    main()
